"""Parameter optimization."""

# Authors: Carolina Fernandez <cxf418@miami.edu>
#          Nick Tolley <nicholas_tolley@brown.edu>
#          Ryan Thorpe <ryan_thorpe@brown.edu>
#          Mainak Jas <mjas@mgh.harvard.edu>

import numpy as np

from .objective_functions import _rmse_evoked, _maximize_psd


class Optimizer:
    def __init__(
        self,
        initial_net,
        tstop,
        constraints,
        set_params,
        solver="bayesian",
        obj_fun="dipole_rmse",
        max_iter=200,
    ):
        """Parameter optimization.

        Parameters
        ----------
        initial_net : instance of Network
            The network object.
        tstop : float
            The simulated dipole's duration.
        constraints : dict
            The user-defined constraints.
        set_params : func
            User-defined function that sets parameters in network drives.

                ``set_params(net, params) -> None``

            where ``net`` is a Network object and ``params`` is a dictionary
            of the parameters that will be set inside the function.
        solver : str
            The optimizer, 'bayesian' or 'cobyla'.
        obj_fun : str | func
            The objective function to be minimized. Can be 'dipole_rmse',
            'maximize_psd', or a user-defined function. The default is
            'dipole_rmse'.
        max_iter : int, optional
            The max number of calls to the objective function. The default is
            200.

        Attributes
        ----------
        constraints : dict
            The user-defined constraints.
        max_iter : int
            The max number of calls to the objective function.
        solver : func
            The optimization function.
        obj_fun : func
            The objective function to be minimized.
        obj_fun_name : str
            The name of the template objective function.
        tstop : float
            The simulated dipole's duration.
        net_ : instance of Network
            The network object with optimized drives.
        obj_ : list
            The objective function values.
        opt_params_ : list
            The list of optimized parameter values.
        """

        if initial_net.external_drives:
            raise ValueError(
                "The current Network instance has external "
                + "drives, provide a Network object with no "
                + "external drives."
            )
        self._initial_net = initial_net
        self.constraints = constraints
        self._set_params = set_params
        self.max_iter = max_iter
        # Optimizer method
        if solver == "bayesian":
            self.solver = "bayesian"
            self._assemble_constraints = _assemble_constraints_bayesian
            self._run_opt = _run_opt_bayesian
        elif solver == "cobyla":
            self.solver = "cobyla"
            self._assemble_constraints = _assemble_constraints_cobyla
            self._run_opt = _run_opt_cobyla
        else:
            raise ValueError("solver must be 'bayesian' or 'cobyla'")
        # Response to be optimized
        if obj_fun == "dipole_rmse":
            self.obj_fun = _rmse_evoked
            self.obj_fun_name = "dipole_rmse"
        elif obj_fun == "maximize_psd":
            self.obj_fun = _maximize_psd
            self.obj_fun_name = "maximize_psd"
        else:
            self.obj_fun = obj_fun  # user-defined function
            self.obj_fun_name = None
        self.tstop = tstop
        self.net_ = None
        self.obj_ = list()
        self.opt_params_ = None

    def __repr__(self):
        is_fit = False
        if self.net_ is not None:
            is_fit = True

        name = self.__class__.__name__
        return f"<{name}\nsolver={self.solver}\nfit={is_fit}>"

    def fit(self, **obj_fun_kwargs):
        """Runs optimization routine.

        Parameters
        ----------
        target : instance of Dipole (if obj_fun='dipole_rmse')
            A dipole object with experimental data.
        f_bands : list of tuples (if obj_fun='maximize_psd')
            Lower and higher limit for each frequency band.
        relative_bandpower : tuple (if obj_fun='maximize_psd')
            Weight for each frequency band.
        scale_factor : float, optional
            The dipole scale factor.
        smooth_window_len : float, optional
            The smooth window length.
        """
        if self.obj_fun_name == "dipole_rmse" and "target" not in obj_fun_kwargs:
            raise Exception("target must be specified")
        elif self.obj_fun_name == "maximize_psd" and (
            "f_bands" not in obj_fun_kwargs
            or "relative_bandpower" not in obj_fun_kwargs
        ):
            raise Exception("f_bands and relative_bandpower must be specified")

        constraints = self._assemble_constraints(self.constraints)
        initial_params = _get_initial_params(self.constraints)

        opt_params, obj, net_ = self._run_opt(
            self._initial_net,
            self.tstop,
            constraints,
            self._set_params,
            self.obj_fun,
            initial_params,
            self.max_iter,
            obj_fun_kwargs,
        )

        self.net_ = net_
        self.obj_ = obj
        self.opt_params_ = opt_params

    def plot_convergence(self, ax=None, show=True):
        """Convergence plot.

        Parameters
        ----------
        ax : instance of matplotlib figure, optional
            The matplotlib axis. The default is None.
        show : bool
            If True, show the figure. The default is True.

        Returns
        -------
        fig : instance of plt.fig
            The matplotlib figure handle.
        """

        import matplotlib as mpl
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots(constrained_layout=True)

        axis = ax if isinstance(ax, mpl.axes._axes.Axes) else ax

        x = list(range(1, self.max_iter + 1))
        y_min = min(self.obj_) - 0.01
        y_max = max(self.obj_) + 0.01

        axis.plot(x, self.obj_, color="black")
        axis.set_ylim([y_min, y_max])
        axis.set_title("Convergence")
        axis.set_xlabel("Number of calls")
        axis.set_ylabel("Objective value")
        axis.grid(visible=True)

        fig.show(show)
        return axis.get_figure()


def _get_initial_params(constraints):
    """Gets initial parameters as midpoints of parameter ranges.

    Parameters
    ----------
    constraints : dict
        The user-defined constraints.

    Returns
    -------
    initial_params : dict
        Keys are parameter names, values are initial parameters.
    """

    initial_params = dict()
    for cons_key in constraints:
        initial_params.update(
            {cons_key: (constraints[cons_key][0] + constraints[cons_key][1]) / 2}
        )

    return initial_params


def _assemble_constraints_bayesian(constraints):
    """Assembles constraints in format required by gp_minimize.

    Parameters
    ----------
    constraints : dict
        The user-defined constraints.

    Returns
    -------
    cons_bayesian : list of tuples
        Lower and higher limit for each parameter.
    """

    # assemble constraints in solver-specific format
    cons_bayesian = list(constraints.values())
    return cons_bayesian


def _assemble_constraints_cobyla(constraints):
    """Assembles constraints in format required by fmin_cobyla.

    Parameters
    ----------
    constraints : dict
        The user-defined constraints.

    Returns
    -------
    cons_bayesian : dict
        Set of functions.
    """

    # assemble constraints in solver-specific format
    cons_cobyla = list()
    for idx, cons_key in enumerate(constraints):
        cons_cobyla.append(lambda x, idx=idx: float(constraints[cons_key][1]) - x[idx])
        cons_cobyla.append(lambda x, idx=idx: x[idx] - float(constraints[cons_key][0]))

    return cons_cobyla


def _update_params(initial_params, predicted_params):
    """Update params with predicted parameters.

    Parameters
    ----------
    initial_params : dict
        Keys are parameter names, values are initial parameters.
    predicted_params : list
        Parameters selected by the optimizer.

    Returns
    -------
    params : dict
        Keys are parameter names, values are parameters.
    """

    params = dict()
    for param_key, param_name in enumerate(initial_params):
        params.update({param_name: predicted_params[param_key]})

    return params


def _run_opt_bayesian(
    initial_net,
    tstop,
    constraints,
    set_params,
    obj_fun,
    initial_params,
    max_iter,
    obj_fun_kwargs,
):
    """Runs optimization routine with gp_minimize optimizer.

    Parameters
    ----------
    initial_net : instance of Network
        The network object.
    tstop : float
        The simulated dipole's duration.
    constraints : list of tuples
        Parameter constraints in solver-specific format.
    set_params : func
        User-defined function that sets parameters in network drives.
    obj_fun : func
        The objective function.
    initial_params : dict
        Keys are parameter names, values are initial parameters.
    max_iter : int
        Number of calls the optimizer makes.

    Returns
    -------
    opt_params : list
        Optimized parameters.
    obj : list
        Objective values.
    net_ : instance of Network
        Optimized network object.
    """

    from ..externals.bayesopt import bayes_opt, expected_improvement

    obj_values = list()

    def _obj_func(predicted_params):
        return obj_fun(
            initial_net=initial_net,
            initial_params=initial_params,
            set_params=set_params,
            predicted_params=predicted_params,
            update_params=_update_params,
            obj_values=obj_values,
            tstop=tstop,
            obj_fun_kwargs=obj_fun_kwargs,
        )

    opt_results, _ = bayes_opt(
        func=_obj_func,
        x0=list(initial_params.values()),
        cons=constraints,
        acquisition=expected_improvement,
        maxfun=max_iter,
    )

    # get optimized params
    opt_params = opt_results

    # get objective values
    obj = [np.min(obj_values[:idx]) for idx in range(1, max_iter + 1)]

    # get optimized net
    params = _update_params(initial_params, opt_params)
    net_ = initial_net.copy()
    set_params(net_, params)

    return opt_params, obj, net_


def _run_opt_cobyla(
    initial_net,
    tstop,
    constraints,
    set_params,
    obj_fun,
    initial_params,
    max_iter,
    obj_fun_kwargs,
):
    """Runs optimization routine with fmin_cobyla optimizer.

    Parameters
    ----------
    initial_net : instance of Network
        The network object.
    tstop : float
        The simulated dipole's duration.
    constraints : list of tuples
        Parameter constraints in solver-specific format.
    set_params : func
        User-defined function that sets parameters in network drives.
    obj_fun : func
        The objective function.
    initial_params : dict
        Keys are parameter names, values are initial parameters.
    max_iter : int
        Number of calls the optimizer makes.

    Returns
    -------
    opt_params : list
        Optimized parameters.
    obj : list
        Objective values.
    net_ : instance of Network
        Optimized network object.
    """

    from scipy.optimize import fmin_cobyla

    obj_values = list()

    def _obj_func(predicted_params):
        return obj_fun(
            initial_net=initial_net,
            initial_params=initial_params,
            set_params=set_params,
            predicted_params=predicted_params,
            update_params=_update_params,
            obj_values=obj_values,
            tstop=tstop,
            obj_fun_kwargs=obj_fun_kwargs,
        )

    opt_results = fmin_cobyla(
        _obj_func,
        cons=constraints,
        rhobeg=0.1,
        rhoend=1e-4,
        x0=list(initial_params.values()),
        maxfun=max_iter,
        catol=0.0,
    )

    # get optimized params
    opt_params = opt_results

    # get objective values
    obj = [np.min(obj_values[:idx]) for idx in range(1, max_iter + 1)]

    # get optimized net
    params = _update_params(initial_params, opt_params)
    net_ = initial_net.copy()
    set_params(net_, params)

    return opt_params, obj, net_



"""Input check functions."""

# Authors: Nick Tolley <nicholas_tolley@brown.edu>

from .params import _long_name
from .externals.mne import _validate_type, _check_option


def _check_gids(gids, gid_ranges, valid_cells, arg_name, same_type=True):
    """Format different gid specifications into list of gids"""
    _validate_type(
        gids, (int, list, range, str, None), arg_name, "int list, range, str, or None"
    )

    # Convert gids to list
    if gids is None:
        return list()
    if isinstance(gids, int):
        gids = [gids]
    elif isinstance(gids, str):
        _check_option(arg_name, gids, valid_cells)
        gids = gid_ranges[_long_name(gids)]

    if all(isinstance(gid, str) for gid in gids):
        gids = [gid for cell_type in gids for gid in gid_ranges[cell_type]]

    cell_type = _gid_to_type(gids[0], gid_ranges)
    for gid in gids:
        _validate_type(gid, int, arg_name)
        gid_type = _gid_to_type(gid, gid_ranges)
        if gid_type is None:
            raise AssertionError(f"{arg_name} {gid} not in net.gid_ranges")
        if same_type and gid_type != cell_type:
            raise AssertionError(f"All {arg_name} must be of the same type")

    return gids


def _gid_to_type(gid, gid_ranges):
    """Reverse lookup of gid to type."""
    for gidtype, gids in gid_ranges.items():
        if gid in gids:
            return gidtype


def _string_input_to_list(input_str, valid_str, arg_name):
    """Convert input strings to list"""
    if input_str is None:
        input_str = list()
    elif isinstance(input_str, str):
        input_str = [input_str]
    for str_item in input_str:
        _check_option(arg_name, str_item, valid_str)

    return input_str



"""
====================
07. Batch Simulation
====================

This example shows how to do batch simulations in HNN-core, allowing users to
efficiently run multiple simulations with different parameters
for comprehensive analysis.
"""

###############################################################################
# Note that batch simulation requires you to install HNN-core with Joblib
# parallel support, which you can do by installing it with
# ``pip install "hnn_core[parallel]"``


# Authors: Abdul Samad Siddiqui <abdulsamadsid1@gmail.com>
#          Nick Tolley <nicholas_tolley@brown.edu>
#          Ryan Thorpe <ryan_thorpe@brown.edu>
#          Mainak Jas <mjas@mgh.harvard.edu>
#
# This project was supported by Google Summer of Code (GSoC) 2024.

###############################################################################
# Let us import ``hnn_core``.

import matplotlib.pyplot as plt
import numpy as np

from hnn_core.batch_simulate import BatchSimulate
from hnn_core import jones_2009_model

# The number of cores may need modifying depending on your current machine.
n_jobs = 4
###############################################################################
# The `add_evoked_drive` function simulates external input to the network,
# mimicking sensory stimulation or other external events.
#
# - `evprox` indicates a proximal drive, targeting dendrites near the cell
#   bodies.
# - `mu=40` and `sigma=5` define the timing (mean and spread) of the input.
# - `weights_ampa` and `synaptic_delays` control the strength and
#   timing of the input.
#
# This evoked drive causes the initial positive deflection in the dipole
# signal, triggering a cascade of activity through the network and
# resulting in the complex waveforms observed.


def set_params(param_values, net=None):
    """
    Set parameters for the network drives.

    Parameters
    ----------
    param_values : dict
        Dictionary of parameter values.
    net : instance of Network, optional
        If None, a new network is created using the specified model type.
    """
    weights_ampa = {'L2_basket': param_values['weight_basket'],
                    'L2_pyramidal': param_values['weight_pyr'],
                    'L5_basket': param_values['weight_basket'],
                    'L5_pyramidal': param_values['weight_pyr']}

    synaptic_delays = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                       'L5_basket': 1., 'L5_pyramidal': 1.}

    # Add an evoked drive to the network.
    net.add_evoked_drive('evprox',
                         mu=40,
                         sigma=5,
                         numspikes=1,
                         location='proximal',
                         weights_ampa=weights_ampa,
                         synaptic_delays=synaptic_delays)

###############################################################################
# Next, we define a parameter grid for the batch simulation.


param_grid = {
    'weight_basket': np.logspace(-4, -1, 20),
    'weight_pyr': np.logspace(-4, -1, 20)
}

###############################################################################
# We then define a function to calculate summary statistics.


def summary_func(results):
    """
    Calculate the min and max dipole peak for each simulation result.

    Parameters
    ----------
    results : list
        List of dictionaries containing simulation results.

    Returns
    -------
    summary_stats : list
        Summary statistics for each simulation result.
    """
    summary_stats = []
    for result in results:
        dpl_smooth = result['dpl'][0].copy().smooth(window_len=30)
        dpl_data = dpl_smooth.data['agg']
        min_peak = np.min(dpl_data)
        max_peak = np.max(dpl_data)
        summary_stats.append({'min_peak': min_peak, 'max_peak': max_peak})
    return summary_stats

###############################################################################
# Run the batch simulation and collect the results.


# Initialize the network model and run the batch simulation.
net = jones_2009_model(mesh_shape=(3, 3))
batch_simulation = BatchSimulate(net=net,
                                 set_params=set_params,
                                 summary_func=summary_func)
simulation_results = batch_simulation.run(param_grid,
                                          n_jobs=n_jobs,
                                          combinations=False,
                                          backend='loky')

print("Simulation results:", simulation_results)
###############################################################################
# This plot shows an overlay of all smoothed dipole waveforms from the
# batch simulation. Each line represents a different set of synaptic strength
# parameters (`weight_basket`), allowing us to visualize the range of responses
# across the parameter space.
# The colormap represents synaptic strengths, from weaker (purple)
# to stronger (yellow).
#
# As drive strength increases, dipole responses show progressively larger
# amplitudes and more distinct features, reflecting heightened network
# activity. Weak drives (purple lines) produce smaller amplitude signals with
# simpler waveforms, while stronger drives (yellow lines) generate
# larger responses with more pronounced oscillatory features, indicating
# more robust network activity.
#
# The y-axis represents dipole amplitude in nAm (nanoAmpere-meters), which is
# the product of current flow and distance in the neural tissue.
#
# Stronger synaptic connections (yellow lines) generally show larger
# amplitude responses and more pronounced features throughout the simulation.

dpl_waveforms, param_values = [], []
for data_list in simulation_results['simulated_data']:
    for data in data_list:
        dpl_smooth = data['dpl'][0].copy().smooth(window_len=30)
        dpl_waveforms.append(dpl_smooth.data['agg'])
        param_values.append(data['param_values']['weight_basket'])

plt.figure(figsize=(10, 6))
cmap = plt.get_cmap('viridis')
log_param_values = np.log10(param_values)
norm = plt.Normalize(log_param_values.min(), log_param_values.max())

for waveform, log_param in zip(dpl_waveforms, log_param_values):
    color = cmap(norm(log_param))
    plt.plot(waveform, color=color, alpha=0.7, linewidth=2)
plt.title('Overlay of Dipole Waveforms')
plt.xlabel('Time (ms)')
plt.ylabel('Dipole Amplitude (nAm)')
plt.grid(True)
plt.tight_layout()
plt.show()
###############################################################################
# This plot displays the minimum and maximum dipole peaks across
# different synaptic strengths. This allows us to see how the range of
# dipole activity changes as we vary the synaptic strength parameter.

min_peaks, max_peaks, param_values = [], [], []
for summary_list, data_list in zip(simulation_results['summary_statistics'],
                                   simulation_results['simulated_data']):
    for summary, data in zip(summary_list, data_list):
        min_peaks.append(summary['min_peak'])
        max_peaks.append(summary['max_peak'])
        param_values.append(data['param_values']['weight_basket'])

# Plotting
plt.figure(figsize=(10, 6))
plt.plot(param_values, min_peaks, label='Min Dipole Peak')
plt.plot(param_values, max_peaks, label='Max Dipole Peak')
plt.xlabel('Synaptic Strength (nS)')
plt.ylabel('Dipole Peak Magnitude')
plt.title('Min and Max Dipole Peaks across Simulations')
plt.legend()
plt.grid(True)
plt.xscale('log')
plt.tight_layout()
plt.show()



"""Handling of parameters."""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>

import json
import fnmatch
import os.path as op
from pathlib import Path
from copy import deepcopy

from .params_default import get_params_default
from .externals.mne import _validate_type


# return number of evoked inputs (proximal, distal)
# using dictionary d (or if d is a string, first load the dictionary from
# filename d)
def _count_evoked_inputs(d):
    nprox = ndist = 0
    for k, _ in d.items():
        if k.startswith("t_"):
            if k.count("evprox") > 0:
                nprox += 1
            elif k.count("evdist") > 0:
                ndist += 1
    return nprox, ndist


def _read_json(param_data):
    """Read param values from a .json file.
    Parameters
    ----------
    param_data : str
        The data read in from the param file

    Returns
    -------
    params_input : dict
        Dictionary of parameters
    """
    return json.loads(param_data)


def _read_legacy_params(param_data):
    """Read param values from a .param file (legacy).
    Parameters
    ----------
    param_data : str
        The data read in from the param file

    Returns
    -------
    params_input : dict
        Dictionary of parameters
    """
    params_input = dict()
    for line in param_data.splitlines():
        split_line = line.lstrip().split(":")
        key, value = [field.strip() for field in split_line]
        try:
            if "." in value or "e" in value:
                params_input[key] = float(value)
            else:
                params_input[key] = int(value)
        except ValueError:
            params_input[key] = str(value)

    return params_input


def read_params(params_fname, file_contents=None):
    """Read param values from a file (.json or .param).

    Parameters
    ----------
    params_fname : str
        Full path to the file (.param)
    file_contents : str | None
        If file_contents are provided as a string,
        it is parsed into a dictionary.

    Returns
    -------
    params : an instance of Params
        Params containing parameter values from file
    """

    split_fname = op.splitext(params_fname)
    ext = split_fname[1]

    if ext not in [".json", ".param"]:
        raise ValueError(
            "Unrecognized extension, expected one of" + " .json, .param. Got %s" % ext
        )

    if file_contents is None:
        with open(params_fname, "r") as fp:
            file_contents = fp.read()

    read_func = {".json": _read_json, ".param": _read_legacy_params}
    params_dict = read_func[ext](file_contents)

    if len(params_dict) == 0:
        raise ValueError(
            "Failed to read parameters from file: %s" % op.normpath(params_fname)
        )

    params = Params(params_dict)

    return params


def _long_name(short_name):
    long_name = dict(
        L2Basket="L2_basket",
        L5Basket="L5_basket",
        L2Pyr="L2_pyramidal",
        L5Pyr="L5_pyramidal",
    )
    if short_name in long_name:
        return long_name[short_name]
    return short_name


def _short_name(short_name):
    long_name = dict(
        L2_basket="L2Basket",
        L5_basket="L5Basket",
        L2_pyramidal="L2Pyr",
        L5_pyramidal="L5Pyr",
    )
    if short_name in long_name:
        return long_name[short_name]
    return short_name


def _extract_bias_specs_from_hnn_params(params, cellname_list):
    """Create 'bias specification' dicts from saved parameters"""
    bias_specs = {"tonic": {}}  # currently only 'tonic' biases known
    for cellname in cellname_list:
        short_name = _short_name(cellname)
        is_tonic_present = [
            f"Itonic_{p}_{short_name}_soma" in params for p in ["A", "t0", "T"]
        ]
        if any(is_tonic_present):
            if not all(is_tonic_present):
                raise ValueError(
                    f"Tonic input must have the amplitude, "
                    f"start time and end time specified. One "
                    f"or more parameter may be missing for "
                    f"cell type {cellname}"
                )
            bias_specs["tonic"][cellname] = {
                "amplitude": params[f"Itonic_A_{short_name}_soma"],
                "t0": params[f"Itonic_t0_{short_name}_soma"],
                "tstop": params[f"Itonic_T_{short_name}_soma"],
            }
    return bias_specs


def _extract_drive_specs_from_hnn_params(params, cellname_list, legacy_mode=False):
    """Create 'drive specification' dicts from saved parameters"""
    # convert legacy params-dict to legacy "feeds" dicts
    p_common, p_unique = create_pext(params, params["tstop"])

    # Using 'feed' for legacy compatibility, 'drives' for new API
    drive_specs = dict()
    for ic, par in enumerate(p_common):
        if (not legacy_mode) and par["tstop"] < par["t0"]:
            continue
        feed_name = f"bursty{ic + 1}"
        drive = dict()
        drive["type"] = "bursty"
        drive["cell_specific"] = False
        drive["dynamics"] = {
            "tstart": par["t0"],
            "tstart_std": par["t0_stdev"],
            "tstop": par["tstop"],
            "burst_rate": par["f_input"],
            "burst_std": par["stdev"],
            "numspikes": par["events_per_cycle"],
            "n_drive_cells": par["n_drive_cells"],
            "spike_isi": 10,
        }  # not exposed in params-files
        drive["location"] = par["loc"]
        drive["space_constant"] = par["lamtha"]
        drive["event_seed"] = par["prng_seedcore"]
        drive["weights_ampa"] = dict()
        drive["weights_nmda"] = dict()
        drive["synaptic_delays"] = dict()

        for cellname in cellname_list:
            cname_ampa = _short_name(cellname) + "_ampa"
            cname_nmda = _short_name(cellname) + "_nmda"
            if cname_ampa in par:
                ampa_weight = par[cname_ampa][0]
                ampa_delay = par[cname_ampa][1]
                drive["weights_ampa"][cellname] = ampa_weight

                # NB synaptic delay same for NMDA, read only for AMPA
                drive["synaptic_delays"][cellname] = ampa_delay

            if cname_nmda in par:
                nmda_weight = par[cname_nmda][0]
                drive["weights_nmda"][cellname] = nmda_weight

        drive_specs[feed_name] = drive

    for feed_name, par in p_unique.items():
        drive = dict()
        drive["cell_specific"] = True
        drive["weights_ampa"] = dict()
        drive["weights_nmda"] = dict()
        drive["synaptic_delays"] = dict()

        if feed_name.startswith("evprox") or feed_name.startswith("evdist"):
            drive["type"] = "evoked"
            if feed_name.startswith("evprox"):
                drive["location"] = "proximal"
            else:
                drive["location"] = "distal"

            cell_keys_present = [key for key in par if key in cellname_list]
            sigma = par[cell_keys_present[0]][3]  # IID for all cells!

            n_drive_cells = "n_cells"
            if par["sync_evinput"]:
                n_drive_cells = 1
                drive["cell_specific"] = False

            drive["dynamics"] = {
                "mu": par["t0"],
                "sigma": sigma,
                "numspikes": par["numspikes"],
                "n_drive_cells": n_drive_cells,
            }
            drive["space_constant"] = par["lamtha"]
            drive["event_seed"] = par["prng_seedcore"]
            # XXX Force random states to be the same as HNN-gui for the default
            # parameter set after increasing the number of bursty drive
            # gids from 2 to 20
            if legacy_mode:
                drive["event_seed"] -= 18
            for cellname in cellname_list:
                if cellname in par:
                    ampa_weight = par[cellname][0]
                    nmda_weight = par[cellname][1]
                    synaptic_delays = par[cellname][2]
                    drive["weights_ampa"][cellname] = ampa_weight
                    drive["weights_nmda"][cellname] = nmda_weight
                    drive["synaptic_delays"][cellname] = synaptic_delays

        # Skip drive if not in legacy mode
        elif feed_name.startswith("extgauss"):
            if (not legacy_mode) and par["L2_basket"][3] > params["tstop"]:
                continue
            drive["type"] = "gaussian"
            drive["location"] = par["loc"]

            drive["dynamics"] = {
                "mu": par["L2_basket"][3],  # NB IID
                "sigma": par["L2_basket"][4],
                "numspikes": 50,  # NB hard-coded in GUI!
                "sync_within_trial": False,
            }
            drive["space_constant"] = par["lamtha"]
            drive["event_seed"] = par["prng_seedcore"]

            for cellname in cellname_list:
                if cellname in par:
                    ampa_weight = par[cellname][0]
                    synaptic_delays = par[cellname][3]
                    drive["weights_ampa"][cellname] = ampa_weight
                    drive["synaptic_delays"][cellname] = synaptic_delays

            drive["weights_nmda"] = dict()  # no NMDA weights for Gaussians
        elif feed_name.startswith("extpois"):
            if (not legacy_mode) and par["t_interval"][1] < par["t_interval"][0]:
                continue
            drive["type"] = "poisson"
            drive["location"] = par["loc"]
            drive["space_constant"] = par["lamtha"]
            drive["event_seed"] = par["prng_seedcore"]

            rate_params = dict()
            for cellname in cellname_list:
                if cellname in par:
                    rate_params[cellname] = par[cellname][3]
                    # XXX correct for non-positive poisson rate constant that
                    # is specified in null poisson drives of legacy
                    # param files
                    if not rate_params[cellname] > 0:
                        rate_params[cellname] = 1
                    ampa_weight = par[cellname][0]
                    nmda_weight = par[cellname][1]
                    synaptic_delays = par[cellname][2]
                    drive["weights_ampa"][cellname] = ampa_weight
                    drive["weights_nmda"][cellname] = nmda_weight
                    drive["synaptic_delays"][cellname] = synaptic_delays

            # do NOT allow negative times sometimes used in param-files
            drive["dynamics"] = {
                "tstart": max(0, par["t_interval"][0]),
                "tstop": max(0, par["t_interval"][1]),
                "rate_constant": rate_params,
            }

        drive_specs[feed_name] = drive
    return drive_specs


class Params(dict):
    """Params object.

    Parameters
    ----------
    params_input : dict | None
        Dictionary of parameters. If None, use default parameters.
    """

    def __init__(self, params_input=None):
        if params_input is None:
            params_input = dict()

        if isinstance(params_input, dict):
            nprox, ndist = _count_evoked_inputs(params_input)
            # create default params templated from params_input
            params_default = get_params_default(nprox, ndist)

            for key in params_default.keys():
                if key in params_input:
                    self[key] = params_input[key]
                else:
                    self[key] = params_default[key]
        else:
            raise ValueError(
                "params_input must be dict or None. Got %s" % type(params_input)
            )

    def __repr__(self):
        """Display the params nicely."""
        return json.dumps(self, sort_keys=True, indent=4)

    def __getitem__(self, key):
        """Return a subset of parameters."""
        keys = self.keys()
        if key in keys:
            return dict.__getitem__(self, key)
        else:
            matches = fnmatch.filter(keys, key)
            if len(matches) == 0:
                return dict.__getitem__(self, key)
            params = self.copy()
            for key in keys:
                if key not in matches:
                    params.pop(key)
            return params

    def __setitem__(self, key, value):
        """Set the value for a subset of parameters."""
        keys = self.keys()
        if key in keys:
            return dict.__setitem__(self, key, value)
        else:
            matches = fnmatch.filter(keys, key)
            if len(matches) == 0:
                return dict.__setitem__(self, key, value)
            for key in keys:
                if key in matches:
                    self.update({key: value})

    def copy(self):
        return deepcopy(self)

    def write(self, fname):
        """Write param values to a file.

        Parameters
        ----------
        fname : str
            Full path to the output file (.json)
        """
        with open(fname, "w") as fp:
            json.dump(self, fp)


def _validate_feed(p_ext_d, tstop):
    """Validate external inputs that are fed to all
    cells uniformly (i.e., rather than individually).
    For now, this only includes rhythmic inputs.

    Parameters
    ----------
    p_ext_d : dict
        The parameter set to validate and append to p_ext.
    tstop : float
        Stop time of the simulation.

    Returns
    -------
    p_ext : list
        Cumulative list of dicts with newly appended ExtFeed.
    """

    # # reset tstop if the specified tstop exceeds the
    # # simulation runtime

    if p_ext_d["tstop"] > tstop:
        p_ext_d["tstop"] = tstop

    # if stdev is zero, increase synaptic weights 5 fold to make
    # single input equivalent to 5 simultaneous input to prevent spiking
    # <<---- SN: WHAT IS THIS RULE!?!?!?
    if not p_ext_d["stdev"]:
        for key in p_ext_d.keys():
            if key.endswith("Pyr"):
                p_ext_d[key] = (p_ext_d[key][0] * 5.0, p_ext_d[key][1])
            elif key.endswith("Basket"):
                p_ext_d[key] = (p_ext_d[key][0] * 5.0, p_ext_d[key][1])

    # if L5 delay is -1, use same delays as L2 unless L2 delay is 0.1 in
    # which case use 1. <<---- SN: WHAT IS THIS RULE!?!?!?
    if p_ext_d["L5Pyr_ampa"][1] == -1:
        for key in p_ext_d.keys():
            if key.startswith("L5"):
                if p_ext_d["L2Pyr"][1] != 0.1:
                    p_ext_d[key] = (p_ext_d[key][0], p_ext_d["L2Pyr"][1])
                else:
                    p_ext_d[key] = (p_ext_d[key][0], 1.0)

    return p_ext_d


def check_evoked_synkeys(p, nprox, ndist):
    # make sure ampa,nmda gbar values are in the param dict for evoked
    # inputs(for backwards compatibility)
    # evoked distal target cell types
    lctprox = ["L2Pyr", "L5Pyr", "L2Basket", "L5Basket"]
    # evoked proximal target cell types
    lctdist = ["L2Pyr", "L5Pyr", "L2Basket"]
    lsy = ["ampa", "nmda"]  # synapse types used in evoked inputs
    for nev, pref, lct in zip(
        [nprox, ndist], ["evprox_", "evdist_"], [lctprox, lctdist]
    ):
        for i in range(nev):
            skey = pref + str(i + 1)
            for sy in lsy:
                for ct in lct:
                    k = "gbar_" + skey + "_" + ct + "_" + sy
                    # if the synapse-specific gbar not present, use the
                    # existing weight for both ampa,nmda
                    if k not in p:
                        p[k] = p["gbar_" + skey + "_" + ct]


#


def check_pois_synkeys(p):
    # make sure ampa,nmda gbar values are in the param dict for Poisson inputs
    # (for backwards compatibility)
    lct = ["L2Pyr", "L5Pyr", "L2Basket", "L5Basket"]  # target cell types
    lsy = ["ampa", "nmda"]  # synapse types used in Poisson inputs
    for ct in lct:
        for sy in lsy:
            k = ct + "_Pois_A_weight_" + sy
            # if the synapse-specific weight not present, set it to 0 in p
            if k not in p:
                p[k] = 0.0


# creates the external feed params based on individual simulation params p


def create_pext(p, tstop):
    """Indexable Python list of param dicts for parallel.

    Turn off individual feeds by commenting out relevant line here.
    always valid, no matter the length.

    Parameters
    ----------
    p : dict
        The parameters returned by ExpParams(f_psim).return_pdict()
    """
    p_common = list()

    # p_unique is a dict of input param types that end up going to each cell
    # uniquely
    p_unique = dict()

    # default params for common proximal inputs
    feed_prox = {
        "f_input": p["f_input_prox"],
        "t0": p["t0_input_prox"],
        "tstop": p["tstop_input_prox"],
        "stdev": p["f_stdev_prox"],
        "L2Pyr_ampa": (p["input_prox_A_weight_L2Pyr_ampa"], p["input_prox_A_delay_L2"]),
        "L2Pyr_nmda": (p["input_prox_A_weight_L2Pyr_nmda"], p["input_prox_A_delay_L2"]),
        "L5Pyr_ampa": (p["input_prox_A_weight_L5Pyr_ampa"], p["input_prox_A_delay_L5"]),
        "L5Pyr_nmda": (p["input_prox_A_weight_L5Pyr_nmda"], p["input_prox_A_delay_L5"]),
        "L2Basket_ampa": (
            p["input_prox_A_weight_L2Basket_ampa"],
            p["input_prox_A_delay_L2"],
        ),
        "L2Basket_nmda": (
            p["input_prox_A_weight_L2Basket_nmda"],
            p["input_prox_A_delay_L2"],
        ),
        "L5Basket_ampa": (
            p["input_prox_A_weight_L5Basket_ampa"],
            p["input_prox_A_delay_L5"],
        ),
        "L5Basket_nmda": (
            p["input_prox_A_weight_L5Basket_nmda"],
            p["input_prox_A_delay_L5"],
        ),
        "events_per_cycle": p["events_per_cycle_prox"],
        "prng_seedcore": int(p["prng_seedcore_input_prox"]),
        "lamtha": 100.0,
        "loc": "proximal",
        "n_drive_cells": p["repeats_prox"],
        "t0_stdev": p["t0_input_stdev_prox"],
        "threshold": p["threshold"],
    }

    # ensures time interval makes sense
    p_common.append(_validate_feed(feed_prox, tstop))

    # default params for common distal inputs
    feed_dist = {
        "f_input": p["f_input_dist"],
        "t0": p["t0_input_dist"],
        "tstop": p["tstop_input_dist"],
        "stdev": p["f_stdev_dist"],
        "L2Pyr_ampa": (p["input_dist_A_weight_L2Pyr_ampa"], p["input_dist_A_delay_L2"]),
        "L2Pyr_nmda": (p["input_dist_A_weight_L2Pyr_nmda"], p["input_dist_A_delay_L2"]),
        "L5Pyr_ampa": (p["input_dist_A_weight_L5Pyr_ampa"], p["input_dist_A_delay_L5"]),
        "L5Pyr_nmda": (p["input_dist_A_weight_L5Pyr_nmda"], p["input_dist_A_delay_L5"]),
        "L2Basket_ampa": (
            p["input_dist_A_weight_L2Basket_ampa"],
            p["input_dist_A_delay_L2"],
        ),
        "L2Basket_nmda": (
            p["input_dist_A_weight_L2Basket_nmda"],
            p["input_dist_A_delay_L2"],
        ),
        "events_per_cycle": p["events_per_cycle_dist"],
        "prng_seedcore": int(p["prng_seedcore_input_dist"]),
        "lamtha": 100.0,
        "loc": "distal",
        "n_drive_cells": p["repeats_dist"],
        "t0_stdev": p["t0_input_stdev_dist"],
        "threshold": p["threshold"],
    }

    p_common.append(_validate_feed(feed_dist, tstop))

    nprox, ndist = _count_evoked_inputs(p)
    # print('nprox,ndist evoked inputs:', nprox, ndist)

    # NEW: make sure all evoked synaptic weights present
    # (for backwards compatibility)
    # could cause differences between output of param files
    # since some nmda weights should be 0 while others > 0

    # XXX dangerzone: params are modified in-place, values are imputed if
    # deemed missing (e.g. if 'gbar_evprox_1_L2Pyr_nmda' is not defined, the
    # code adds it to the p-dict with value: p['gbar_evprox_1_L2Pyr'])
    check_evoked_synkeys(p, nprox, ndist)

    # Create proximal evoked response parameters
    # f_input needs to be defined as 0
    for i in range(nprox):
        skey = "evprox_" + str(i + 1)
        p_unique["evprox" + str(i + 1)] = {
            "t0": p["t_" + skey],
            "L2_pyramidal": (
                p["gbar_" + skey + "_L2Pyr_ampa"],
                p["gbar_" + skey + "_L2Pyr_nmda"],
                0.1,
                p["sigma_t_" + skey],
            ),
            "L2_basket": (
                p["gbar_" + skey + "_L2Basket_ampa"],
                p["gbar_" + skey + "_L2Basket_nmda"],
                0.1,
                p["sigma_t_" + skey],
            ),
            "L5_pyramidal": (
                p["gbar_" + skey + "_L5Pyr_ampa"],
                p["gbar_" + skey + "_L5Pyr_nmda"],
                1.0,
                p["sigma_t_" + skey],
            ),
            "L5_basket": (
                p["gbar_" + skey + "_L5Basket_ampa"],
                p["gbar_" + skey + "_L5Basket_nmda"],
                1.0,
                p["sigma_t_" + skey],
            ),
            "prng_seedcore": int(p["prng_seedcore_" + skey]),
            "lamtha": 3.0,
            "loc": "proximal",
            "sync_evinput": p["sync_evinput"],
            "threshold": p["threshold"],
            "numspikes": p["numspikes_" + skey],
        }

    # Create distal evoked response parameters
    # f_input needs to be defined as 0
    for i in range(ndist):
        skey = "evdist_" + str(i + 1)
        p_unique["evdist" + str(i + 1)] = {
            "t0": p["t_" + skey],
            "L2_pyramidal": (
                p["gbar_" + skey + "_L2Pyr_ampa"],
                p["gbar_" + skey + "_L2Pyr_nmda"],
                0.1,
                p["sigma_t_" + skey],
            ),
            "L5_pyramidal": (
                p["gbar_" + skey + "_L5Pyr_ampa"],
                p["gbar_" + skey + "_L5Pyr_nmda"],
                0.1,
                p["sigma_t_" + skey],
            ),
            "L2_basket": (
                p["gbar_" + skey + "_L2Basket_ampa"],
                p["gbar_" + skey + "_L2Basket_nmda"],
                0.1,
                p["sigma_t_" + skey],
            ),
            "prng_seedcore": int(p["prng_seedcore_" + skey]),
            "lamtha": 3.0,
            "loc": "distal",
            "sync_evinput": p["sync_evinput"],
            "threshold": p["threshold"],
            "numspikes": p["numspikes_" + skey],
        }

    # this needs to create many feeds
    # (amplitude, delay, mu, sigma). ordered this way to preserve compatibility
    # NEW: note double weight specification since only use ampa for gauss
    # inputs
    p_unique["extgauss"] = {
        "stim": "gaussian",
        "L2_basket": (
            p["L2Basket_Gauss_A_weight"],
            p["L2Basket_Gauss_A_weight"],
            1.0,
            p["L2Basket_Gauss_mu"],
            p["L2Basket_Gauss_sigma"],
        ),
        "L2_pyramidal": (
            p["L2Pyr_Gauss_A_weight"],
            p["L2Pyr_Gauss_A_weight"],
            0.1,
            p["L2Pyr_Gauss_mu"],
            p["L2Pyr_Gauss_sigma"],
        ),
        "L5_basket": (
            p["L5Basket_Gauss_A_weight"],
            p["L5Basket_Gauss_A_weight"],
            1.0,
            p["L5Basket_Gauss_mu"],
            p["L5Basket_Gauss_sigma"],
        ),
        "L5_pyramidal": (
            p["L5Pyr_Gauss_A_weight"],
            p["L5Pyr_Gauss_A_weight"],
            1.0,
            p["L5Pyr_Gauss_mu"],
            p["L5Pyr_Gauss_sigma"],
        ),
        "lamtha": 100.0,
        "prng_seedcore": int(p["prng_seedcore_extgauss"]),
        "loc": "proximal",
        "threshold": p["threshold"],
    }

    check_pois_synkeys(p)

    # Poisson distributed inputs to proximal
    # NEW: setting up AMPA and NMDA for Poisson inputs; why delays differ?
    p_unique["extpois"] = {
        "stim": "poisson",
        "L2_basket": (
            p["L2Basket_Pois_A_weight_ampa"],
            p["L2Basket_Pois_A_weight_nmda"],
            1.0,
            p["L2Basket_Pois_lamtha"],
        ),
        "L2_pyramidal": (
            p["L2Pyr_Pois_A_weight_ampa"],
            p["L2Pyr_Pois_A_weight_nmda"],
            0.1,
            p["L2Pyr_Pois_lamtha"],
        ),
        "L5_basket": (
            p["L5Basket_Pois_A_weight_ampa"],
            p["L5Basket_Pois_A_weight_nmda"],
            1.0,
            p["L5Basket_Pois_lamtha"],
        ),
        "L5_pyramidal": (
            p["L5Pyr_Pois_A_weight_ampa"],
            p["L5Pyr_Pois_A_weight_nmda"],
            1.0,
            p["L5Pyr_Pois_lamtha"],
        ),
        "lamtha": 100.0,
        "prng_seedcore": int(p["prng_seedcore_extpois"]),
        "t_interval": (p["t0_pois"], p["T_pois"]),
        "loc": "proximal",
        "threshold": p["threshold"],
    }

    return p_common, p_unique


# Takes two dictionaries (d1 and d2) and compares the keys in d1 to those in d2
# if any match, updates the (key, value) pair of d1 to match that of d2
# not real happy with variable names, but will have to do for now
def compare_dictionaries(d1, d2):
    # iterate over intersection of key sets (i.e. any common keys)
    for key in d1.keys() and d2.keys():
        # update d1 to have same (key, value) pair as d2
        d1[key] = d2[key]

    return d1


def _any_positive_weights(drive):
    """Checks a drive for any positive weights."""
    weights = list(drive["weights_ampa"].values()) + list(
        drive["weights_nmda"].values()
    )
    if any([val > 0 for val in weights]):
        return True
    else:
        return False


def remove_nulled_drives(net):
    """Removes drives from network if they have been given null parameters.

    Legacy param files contained parameter placeholders for non-functional
    drives. These drives were nulled by assigning values outside typical
    ranges. This function removes drives on the following conditions:
        1. Start time is larger than stop time
        2. All weights are non-positive

    Parameters
    ----------
    net : Network object

    Returns
    -------
    net : Network object

    """
    from .network import pick_connection

    net = deepcopy(net)
    drives_copy = net.external_drives.copy()

    extras = dict()
    for drive_name, drive in net.external_drives.items():
        conn_indices = pick_connection(net, src_gids=drive_name)

        space_constant = net.connectivity[conn_indices[0]]["nc_dict"]["lamtha"]
        probability = net.connectivity[conn_indices[0]]["probability"]

        extras[drive_name] = {
            "space_constant": space_constant,
            "probability": probability,
        }

    net.clear_drives()
    for drive_name, drive in drives_copy.items():
        # Do not add drive if tstart is > tstop, or negative
        t_start = drive["dynamics"].get("tstart")
        t_stop = drive["dynamics"].get("tstop")
        if (
            t_start is not None
            and t_stop is not None
            and ((t_start > t_stop) or (t_start < 0) or (t_stop < 0))
        ):
            continue
        # Do not add if all 0 weights
        elif not _any_positive_weights(drive):
            continue
        else:
            # Set n_drive_cells to 'n_cells' if equal to max number of cells
            if drive["cell_specific"]:
                drive["n_drive_cells"] = "n_cells"
            net._attach_drive(
                drive["name"],
                drive,
                drive["weights_ampa"],
                drive["weights_nmda"],
                drive["location"],
                extras[drive_name]["space_constant"],
                drive["synaptic_delays"],
                drive["n_drive_cells"],
                drive["cell_specific"],
                extras[drive_name]["probability"],
            )
    return net


def convert_to_json(params_fname, out_fname, include_drives=True, overwrite=True):
    """Converts legacy json or param format to hierarchical json format

    Parameters
    ----------
    params_fname : str or Path
        Path to file
    out_fname: str
        Path to output
    include_drives: bool, default=True
        Include drives from params file
    overwrite: bool, default=True
        Overwrite file
    Returns
    -------
    None
    """
    from .network_models import jones_2009_model

    # Validate inputs
    _validate_type(params_fname, (str, Path), "params_fname")
    _validate_type(out_fname, (str, Path), "out_fname")
    # Convert to Path
    params_fname = Path(params_fname)
    out_fname = Path(out_fname)
    params_suffix = params_fname.suffix.lower().split(".")[-1]

    # Add suffix if not supplied
    if out_fname.suffix != ".json":
        out_fname = out_fname.with_suffix(".json")

    net = jones_2009_model(
        params=read_params(params_fname),
        add_drives_from_params=include_drives,
        legacy_mode=(True if params_suffix == "param" else False),
    )

    # Remove drives that have null attributes
    net = remove_nulled_drives(net)

    net.write_configuration(
        fname=out_fname,
        overwrite=overwrite,
    )
    return


# debug test function
if __name__ == "__main__":
    fparam = "param/debug.param"



"""Visualization functions."""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>
#          Christopher Bailey <cjb@cfin.au.dk>

import numpy as np
from itertools import cycle
import colorsys
import warnings
from .externals.mne import _validate_type


def _lighten_color(color, amount=0.5):
    import matplotlib.colors as mc

    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


def _get_plot_data_trange(times, data, tmin=None, tmax=None):
    """Get slices of times and data based on tmin and tmax"""
    if isinstance(times, list):
        times = np.array(times)
    if isinstance(data, list):
        data = np.array(data)
    plot_tmin = times[0]
    if tmin is not None:
        plot_tmin = max(tmin, plot_tmin)
    plot_tmax = times[-1]
    if tmax is not None:
        plot_tmax = min(tmax, plot_tmax)

    mask = np.logical_and(times >= plot_tmin, times < plot_tmax)

    return data[mask], times[mask]


def _decimate_plot_data(decim, data, times, sfreq=None):
    from scipy.signal import decimate

    if not isinstance(decim, list):
        decim = [decim]

    for dec in decim:
        if not isinstance(dec, int) or dec < 1:
            raise ValueError(
                "each decimation factor must be a positive int, "
                f"but {dec} is a {type(dec)}"
            )
        data = decimate(data, dec)
        times = times[::dec]

    if sfreq is None:
        return data, times
    else:
        sfreq /= np.prod(decim)
        return data, times, sfreq


def plt_show(show=True, fig=None, **kwargs):
    """Show a figure while suppressing warnings.

    NB copied from :func:`mne.viz.utils.plt_show`.

    Parameters
    ----------
    show : bool
        Show the figure.
    fig : instance of Figure | None
        If non-None, use fig.show().
    **kwargs : dict
        Extra arguments for :func:`matplotlib.pyplot.show`.
    """
    from matplotlib import get_backend
    import matplotlib.pyplot as plt

    if show and get_backend() != "agg":
        (fig or plt).show(**kwargs)


def plot_laminar_lfp(
    times,
    data,
    contact_labels,
    tmin=None,
    tmax=None,
    ax=None,
    decim=None,
    color="cividis",
    voltage_offset=50,
    voltage_scalebar=200,
    show=True,
):
    """Plot laminar extracellular electrode array voltage time series.

    Parameters
    ----------
    times : array-like, shape (n_times,)
        Sampling times (in ms).
    data : Two-dimensional Numpy array
        The extracellular voltages as an (n_contacts, n_times) array.
    ax : instance of matplotlib figure | None
        The matplotlib axis
    decim : int | list of int | None (default)
        Optional (integer) factor by which to decimate the raw dipole traces.
        The SciPy function :func:`~scipy.signal.decimate` is used, which
        recommends values <13. To achieve higher decimation factors, a list of
        ints can be provided. These are applied successively.
    color : str | array of floats | ``matplotlib.colors.ListedColormap``
        The colormap to use for plotting. The usual Matplotlib standard
        colormap strings may be used (e.g., 'jetblue'). A color can also be
        defined as an RGBA-quadruplet, or an array of RGBA-values (one for each
        electrode contact trace to plot). An instance of
        :class:`~matplotlib.colors.ListedColormap` may also be provided.
    voltage_offset : float | None (optional)
        Amount to offset traces by on the voltage-axis. Useful for plotting
        laminar arrays.
    voltage_scalebar : float | None (optional)
        Height, in units of uV, of a scale bar to plot in the top-left corner
        of the plot.
    contact_labels : list
        Labels associated with the contacts to plot. Passed as-is to
        :func:`~matplotlib.axes.Axes.set_yticklabels`.
    show : bool
        If True, show the figure

    Returns
    -------
    fig : instance of plt.fig
        The matplotlib figure handle into which time series were plotted.
    """
    import matplotlib.pyplot as plt
    from matplotlib.colors import ListedColormap

    _validate_type(times, (list, np.ndarray), "times")
    _validate_type(data, (list, np.ndarray), "data")
    if isinstance(times, list):
        times = np.array(times)
    if isinstance(data, list):
        data = np.array(data)
    if data.ndim != 2:
        raise ValueError(f"data must be 2D, got shape {data.shape}")
    if len(times) != data.shape[1]:
        raise ValueError(
            f"length of times ({len(times)}) and data ({len(data)}) do not match"
        )

    n_contacts = data.shape[0]
    if color is not None:
        _validate_type(color, (str, tuple, list, np.ndarray, ListedColormap), "color")
        if isinstance(color, (tuple, list)):
            if (
                not np.all([isinstance(c, float) for c in color])
                or len(color) < 3
                or len(color) > 4
            ):
                raise ValueError(f"color must be length 3 or 4, got {color}")
        elif isinstance(color, np.ndarray):
            if color.shape[0] != n_contacts or (
                color.shape[1] < 3 or color.shape[1] > 4
            ):
                raise ValueError(f"color must be n_contacts x (3 or 4), got {color}")
        elif isinstance(color, ListedColormap):
            if color.N != n_contacts:
                raise ValueError(
                    f"ListedColormap has N={color.N}, but "
                    f"there are {n_contacts} contacts"
                )
        elif isinstance(color, str):
            color = plt.get_cmap(color, len(contact_labels))

    if ax is None:
        _, ax = plt.subplots(1, 1)

    n_offsets = data.shape[0]
    trace_offsets = np.zeros((n_offsets, 1))
    if voltage_offset is not None:
        trace_offsets = np.arange(n_offsets)[:, np.newaxis] * voltage_offset

    for contact_no, trace in enumerate(np.atleast_2d(data)):
        plot_data = trace
        plot_times = times

        if decim is not None:
            plot_data, plot_times = _decimate_plot_data(decim, plot_data, plot_times)

        if isinstance(color, np.ndarray):
            col = color[contact_no]
        elif isinstance(color, ListedColormap):
            col = color(contact_no)
        else:
            col = color
        ax.plot(
            plot_times,
            plot_data + trace_offsets[contact_no],
            label=f"C{contact_no}",
            color=col,
        )

        # To be removed after deprecation cycle
        if tmin is not None or tmax is not None:
            ax.set_xlim(left=tmin, right=tmax)
            warnings.warn(
                "tmin and tmax are deprecated and will be "
                "removed in future releases of hnn-core. Please"
                "use matplotlib plt.xlim to set tmin and tmax.",
                DeprecationWarning,
            )

        else:
            ax.set_xlim(left=times[0], right=times[-1])
    if voltage_offset is not None:
        ax.set_ylim(-voltage_offset, n_offsets * voltage_offset)
        ylabel = "Individual contact traces"
        if len(contact_labels) != n_offsets:
            raise ValueError(
                f"contact_labels is length {len(contact_labels)},"
                f" but {n_offsets} contacts to be plotted"
            )
        else:
            trace_ticks = np.arange(
                0, len(contact_labels) * voltage_offset, voltage_offset
            )
            ax.set_yticks(trace_ticks)
            ax.set_yticklabels(contact_labels)

        if voltage_scalebar is None:
            voltage_scalebar = voltage_offset

    if voltage_scalebar is not None:
        from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar

        scalebar = AnchoredSizeBar(
            ax.transData,
            1,
            f"{voltage_scalebar:.0f} " + r"$\mu V$",
            "upper left",
            size_vertical=voltage_scalebar,
            pad=0.1,
            color="black",
            label_top=False,
            frameon=False,
        )
        ax.add_artist(scalebar)
    else:
        ylabel = r"Electric potential ($\mu V$)"
        ax.ticklabel_format(axis="both", scilimits=(-2, 3))

    ax.set_ylabel(ylabel, multialignment="center")
    ax.set_xlabel("Time (ms)")

    plt_show(show)
    return ax.get_figure()


def plot_dipole(
    dpl,
    tmin=None,
    tmax=None,
    ax=None,
    layer="agg",
    decim=None,
    color="k",
    label="average",
    average=False,
    show=True,
):
    """Simple layer-specific plot function.

    Parameters
    ----------
    dpl : instance of Dipole | list of Dipole instances
        The Dipole object.
    ax : instance of matplotlib figure | None
        The matplotlib axis
    layer : str
        The layer to plot. Can be one of
        'agg', 'L2', and 'L5'
    decim : int or list of int or None (default)
        Optional (integer) factor by which to decimate the raw dipole traces.
        The SciPy function :func:`~scipy.signal.decimate` is used, which
        recommends values <13. To achieve higher decimation factors, a list of
        ints can be provided. These are applied successively.
    color : tuple of float | str
        RGBA value to use for plotting. By default, 'k' (black)
    label : str
        Dipole label. Enabled when average=True
    average : bool
        If True, render the average across all dpls.
    show : bool
        If True, show the figure

    Returns
    -------
    fig : instance of plt.fig
        The matplotlib figure handle.
    """
    import matplotlib.pyplot as plt
    from .dipole import Dipole, average_dipoles

    layers = layer if isinstance(layer, list) else [layer]
    if ax is None:
        _, ax = plt.subplots(
            len(layers), 1, constrained_layout=True, sharex=True, sharey=True
        )
    axes = ax if isinstance(ax, (list, np.ndarray)) else [ax]

    if isinstance(dpl, Dipole):
        dpl = [dpl]
    for this_dpl in dpl:
        _validate_type(this_dpl, Dipole, "dpl", "Dipole, list of Dipole")

    if average:
        dpl.append(average_dipoles(dpl))

    scale_applied = dpl[0].scale_applied

    assert len(layers) == len(axes), "ax and layer should have the same size"

    for layer, ax in zip(layers, axes):
        for idx, dpl_trial in enumerate(dpl):
            if dpl_trial.scale_applied != scale_applied:
                raise RuntimeError("All dipoles must be scaled equally!")

            if layer in dpl_trial.data.keys():
                # extract scaled data and times
                data = dpl_trial.data[layer]
                times = dpl_trial.times

                if decim is not None:
                    data, times = _decimate_plot_data(decim, data, times)
                if idx == len(dpl) - 1 and average:
                    # the average dpl
                    ax.plot(times, data, color=color, label=label, lw=1.5)
                else:
                    alpha = 0.5 if average else 1.0
                    ax.plot(
                        times,
                        data,
                        color=_lighten_color(color, 0.5),
                        alpha=alpha,
                        lw=1.0,
                    )

            # To be removed after deprecation cycle
            if tmin is not None or tmax is not None:
                if tmin is not None or tmax is not None:
                    warnings.warn(
                        "tmin and tmax are deprecated and will be "
                        "removed in future releases of hnn-core. "
                        "Please use matplotlib plt.xlim to set tmin"
                        " and tmax.",
                        DeprecationWarning,
                    )
                ax.set_xlim(left=tmin, right=tmax)
            else:
                ax.set_xlim(left=0, right=times[-1])
        if average:
            ax.legend()

        ax.ticklabel_format(axis="both", scilimits=(-2, 3))
        ax.set_xlabel("Time (ms)")
        if scale_applied == 1:
            ylabel = "Dipole moment (nAm)"
        else:
            ylabel = "Dipole moment\n(nAm " + r"$\times$ {:.0f})".format(scale_applied)
        ax.set_ylabel(ylabel, multialignment="center")
        if layer == "agg":
            title_str = "Aggregate (L2/3 + L5)"
        elif layer == "L2":
            title_str = "L2/3"
        else:
            title_str = layer
        ax.set_title(title_str)

    plt_show(show)
    return axes[0].get_figure()


def plot_spikes_hist(
    cell_response,
    trial_idx=None,
    ax=None,
    spike_types=None,
    color=None,
    invert_spike_types=None,
    show=True,
    **kwargs_hist,
):
    """Plot the histogram of spiking activity across trials.

    Parameters
    ----------
    cell_response : instance of CellResponse
        The CellResponse object from net.cell_response
    trial_idx : int | list of int | None
        Index of trials to be plotted. If None, all trials plotted.
    ax : instance of matplotlib axis | None
        An axis object from matplotlib. If None,
        a new figure is created.
    spike_types: string | list | dictionary | None
        String input of a valid spike type is plotted individually.

        | Ex: ``'poisson'``, ``'evdist'``, ``'evprox'``, ...

        List of valid string inputs will plot each spike type individually.

        | Ex: ``['poisson', 'evdist']``

        Dictionary of valid lists will plot list elements as a group.

        | Ex: ``{'Evoked': ['evdist', 'evprox'], 'Tonic': ['poisson']}``

        If None, all input spike types are plotted individually if any
        are present. Otherwise spikes from all cells are plotted.
        Valid strings also include leading characters of spike types

        | Ex: ``'ev'`` is equivalent to ``['evdist', 'evprox']``
    invert_spike_types: string | list | None
        String input of a valid spike type to be mirrored about the y axis

        | Ex: ``'evdist'``, ``'evprox'``, ...

        List of valid spike types to be mirrored about the y axis

        | Ex: ``['evdist', 'evprox']``

        If None, all input spike types are plotted on the same y axis
    color : str | list of str | dict | None
        Input defining colors of plotted histograms. If str, all
        histograms plotted with same color. If list of str provided,
        histograms for each spike type will be plotted by cycling
        through colors in the list.

        If dict, colors must be specified for all spike_types as a key.
        If a group of spike types is defined by the `spike_types`
        parameter (see dictionary example for `spike_types`),
        the name of this group must be used to specify the colors.

        | Ex: ``{'evdist': 'g', 'evprox': 'r'}``, ``{'Tonic': 'b'}``

        If None, default color cycle used.
    show : bool
        If True, show the figure.
    **kwargs_hist : dict
        Additional keyword arguments to pass to ax.hist.

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure handle.
    """
    import matplotlib.pyplot as plt

    n_trials = len(cell_response.spike_times)
    if trial_idx is None:
        trial_idx = list(range(n_trials))

    if isinstance(trial_idx, int):
        trial_idx = [trial_idx]
    _validate_type(trial_idx, list, "trial_idx", "int, list of int")

    # Extract desired trials
    if len(cell_response._spike_times[0]) > 0:
        spike_times = np.concatenate(
            np.array(cell_response._spike_times, dtype=object)[trial_idx]
        )
        spike_types_data = np.concatenate(
            np.array(cell_response._spike_types, dtype=object)[trial_idx]
        )
    else:
        spike_times = np.array([])
        spike_types_data = np.array([])

    unique_types = np.unique(spike_types_data)
    spike_types_mask = {
        s_type: np.isin(spike_types_data, s_type) for s_type in unique_types
    }
    cell_types = ["L5_pyramidal", "L5_basket", "L2_pyramidal", "L2_basket"]
    input_types = np.setdiff1d(unique_types, cell_types)

    if isinstance(spike_types, str):
        spike_types = {spike_types: [spike_types]}

    if spike_types is None:
        if any(input_types):
            spike_types = input_types.tolist()
        else:
            spike_types = unique_types.tolist()
    if isinstance(spike_types, list):
        spike_types = {s_type: [s_type] for s_type in spike_types}
    if isinstance(spike_types, dict):
        for spike_label in spike_types:
            if not isinstance(spike_types[spike_label], list):
                raise TypeError(
                    f"spike_types[{spike_label}] must be a list. "
                    f"Got "
                    f"{type(spike_types[spike_label]).__name__}."
                )

    if not isinstance(spike_types, dict):
        raise TypeError("spike_types should be str, list, dict, or None")

    spike_labels = dict()
    for spike_label, spike_type_list in spike_types.items():
        for spike_type in spike_type_list:
            n_found = 0
            for unique_type in unique_types:
                if unique_type.startswith(spike_type):
                    if unique_type in spike_labels:
                        raise ValueError(
                            f"Elements of spike_types must map to"
                            f" mutually exclusive input types."
                            f" {unique_type} is found more than"
                            f" once."
                        )
                    spike_labels[unique_type] = spike_label
                    n_found += 1
            if n_found == 0:
                raise ValueError(f"No input types found for {spike_type}")

    if ax is None:
        _, ax = plt.subplots(1, 1, constrained_layout=True)

    _validate_type(color, (str, list, dict, None), "color", "str, list of str, or dict")

    if color is None:
        color_cycle = cycle(["r", "g", "b", "y", "m", "c"])
    elif isinstance(color, str):
        color_cycle = cycle([color])
    elif isinstance(color, list):
        color_cycle = cycle(color)

    if len(cell_response.times) > 0:
        bins = np.linspace(0, cell_response.times[-1], 50)
    else:
        bins = np.linspace(0, spike_times[-1], 50)

    # Create dictionary to aggregate spike times that have the same spike_label
    spike_type_times = {
        spike_label: list() for spike_label in np.unique(list(spike_labels.values()))
    }
    spike_color = dict()  # Store colors specified for each spike_label
    for spike_type, spike_label in spike_labels.items():
        if spike_label not in spike_color:
            if isinstance(color, dict):
                if spike_label not in color:
                    raise ValueError(
                        f"'{spike_label}' must be defined in color dictionary"
                    )
                _validate_type(
                    color[spike_label], str, "Dictionary values of color", "str"
                )
                spike_color[spike_label] = color[spike_label]
            else:
                spike_color[spike_label] = next(color_cycle)
        spike_type_times[spike_label].extend(spike_times[spike_types_mask[spike_type]])

    if invert_spike_types is None:
        invert_spike_types = list()
    else:
        if not isinstance(invert_spike_types, (str, list)):
            raise TypeError(
                "'invert_spike_types' must be a string or a list of strings"
            )
        if isinstance(invert_spike_types, str):
            invert_spike_types = [invert_spike_types]

        # Check that spike types to invert are correctly specified
        unique_inputs = set(spike_labels.values())
        unique_invert_inputs = set(invert_spike_types)
        check_intersection = unique_invert_inputs.intersection(unique_inputs)
        if not check_intersection == unique_invert_inputs:
            raise ValueError(
                "Elements of 'invert_spike_types' mustmap to valid input types"
            )

    # Initialize secondary axis
    ax1 = None

    # Plot aggregated spike_times
    for spike_label, plot_data in spike_type_times.items():
        hist_color = spike_color[spike_label]

        # Plot on the primary y-axis
        if spike_label not in invert_spike_types:
            ax.hist(plot_data, bins, label=spike_label, color=hist_color, **kwargs_hist)
        # Plot on secondary y-axis
        else:
            if ax1 is None:
                ax1 = ax.twinx()
            ax1.hist(
                plot_data, bins, label=spike_label, color=hist_color, **kwargs_hist
            )
            # Need to add label for easy removal later

    # Set the y-limits based on the maximum across both axes
    if ax1 is not None:
        ax_ylim = ax.get_ylim()[1]
        ax1_ylim = ax1.get_ylim()[1]

        y_max = max(ax_ylim, ax1_ylim)
        ax.set_ylim(0, y_max)
        ax1.set_ylim(0, y_max)
        ax1.invert_yaxis()
        ax1.set_label("Inverted spike histogram")

    if len(cell_response.times) > 0:
        ax.set_xlim(left=0, right=cell_response.times[-1])
    else:
        ax.set_xlim(left=0)

    ax.set_ylabel("Counts")
    ax.set_label("Spike histogram")

    if ax1 is not None:
        # Combine legends
        handles, labels = ax.get_legend_handles_labels()
        handles1, labels1 = ax1.get_legend_handles_labels()
        handles.extend(handles1)
        labels.extend(labels1)

        ax1.legend(handles, labels, loc="upper left")
    else:
        ax.legend()

    plt_show(show)
    return ax.get_figure()


def plot_spikes_raster(
    cell_response,
    trial_idx=None,
    ax=None,
    show=True,
    cell_types=None,
    colors=None,
    show_legend=True,
    marker_size=1.0,
    dpl=None,
    overlay_dipoles=False,
):
    """Plot the aggregate spiking activity according to cell type.

    Parameters
    ----------
    cell_response : instance of CellResponse
        The CellResponse object from net.cell_response
    trial_idx : int | list of int | None
        Index of trials to be plotted. If None, all trials plotted
    ax : instance of matplotlib axis | None
        An axis object from matplotlib. If None, a new figure is created.
    show : bool
        If True, show the figure.
    cell_types : list of str
        List of cell types to plot
    colors : list of str | None
        Optional custom colors to plot. Default will use the color cycler.
    show_legend : bool
        If True, show the legend with colors for cell types
    marker_size : float
        Optional marker size to use when plotting spikes. Uses
        "linelengths" argument of ax.eventplot, which accepts positive
        numeric values only
    dpl : instance of Dipole | list
        The Dipole object containing layer-specific dipole data
        to overlay on the raster plot
    overlay_dipoles : bool
        If True, overlay the layer-specific dipole data on the
        raster plot

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure object.
    """

    import matplotlib.pyplot as plt
    from .dipole import Dipole, average_dipoles

    n_trials = len(cell_response.spike_times)
    if trial_idx is None:
        trial_idx = list(range(n_trials))

    # Get spike types from cell response
    unique_spike_types = cell_response.cell_types

    # validate trial argument
    if isinstance(trial_idx, int):
        trial_idx = [trial_idx]
    _validate_type(trial_idx, list, "trial_idx", "int, list of int")

    # validate cell types
    if cell_types:
        _validate_type(cell_types, list, "cell_types", "list of str")
        if not set(cell_types).issubset(set(unique_spike_types)):
            raise ValueError(
                "Invalid cell types provided. "
                f"Must be of set {unique_spike_types}. "
                f"Got {cell_types}"
            )
    else:
        # Use default cell types
        cell_types = ["L2_basket", "L2_pyramidal", "L5_basket", "L5_pyramidal"]

    # Set default colors
    default_colors = plt.rcParams["axes.prop_cycle"].by_key()["color"][
        : len(cell_types)
    ]
    cell_colors = {cell: color for cell, color in zip(cell_types, default_colors)}

    # validate colors argument
    _validate_type(colors, (list, dict, None), "color", "list of str, or dict")
    if colors:
        if isinstance(colors, list):
            if len(colors) != len(cell_types):
                raise ValueError(
                    f"Number of colors must be equal to number of "
                    f"cell types. {len(colors)} colors provided "
                    f"for {len(cell_types)} cell types."
                )
            cell_colors = {cell: color for cell, color in zip(cell_types, colors)}

        if isinstance(colors, dict):
            # Check valid cell types
            if not set(colors.keys()).issubset(set(unique_spike_types)):
                raise ValueError(
                    "Invalid cell types provided. "
                    f"Must be of set {unique_spike_types}. "
                    f"Got {colors.keys()}"
                )
            cell_colors.update(colors)

    # validate show_legend argument
    _validate_type(show_legend, bool, "show_legend", "bool")

    # validate marker_size
    _validate_type(marker_size, (float, int), "marker_size", "float or int")

    # if marker_size is <= 0, set it to the default value of 1.0
    if marker_size <= 0:
        marker_size = 1.0

    # Extract desired trials
    spike_times = np.concatenate(
        np.array(cell_response._spike_times, dtype=object)[trial_idx]
    )
    spike_types = np.concatenate(
        np.array(cell_response._spike_types, dtype=object)[trial_idx]
    )
    spike_gids = np.concatenate(
        np.array(cell_response._spike_gids, dtype=object)[trial_idx]
    )

    if ax is None:
        _, ax = plt.subplots(1, 1, constrained_layout=True)

    events = []
    for cell_type, color in cell_colors.items():
        cell_type_gids = np.unique(spike_gids[spike_types == cell_type])
        cell_type_times, cell_type_ypos = [], []

        for gid in cell_type_gids:
            gid_time = spike_times[spike_gids == gid]
            cell_type_times.append(gid_time)
            cell_type_ypos.append(-gid)

        if cell_type_times:
            events.append(
                ax.eventplot(
                    cell_type_times,
                    lineoffsets=cell_type_ypos,
                    color=color,
                    label=cell_type,
                    linelengths=marker_size,
                )
            )
        else:
            # Blank plot for no spiking
            events.append(
                ax.eventplot(
                    [-1], lineoffsets=[-1], color=color, label=cell_type, linelengths=1
                )
            )

    # Overlay dipoles on raster plot
    if overlay_dipoles:
        # Confirm that the dpl is not None
        if dpl is None:
            raise ValueError("Dipole object must be provided to overlay dipoles")
        # In the case that a single Dipole object is provided, cast to a list
        if isinstance(dpl, Dipole):
            dpl = [dpl]
        # Validate that all list elements are Dipole objects
        for trial in dpl:
            _validate_type(trial, Dipole, "dpl", "Dipole, list of Dipole")

        # Get (average) layer-sepcific dipole data
        avgs = []
        avgs.append(average_dipoles(dpl))
        l2_dipole = avgs[0].data["L2"]
        l5_dipole = avgs[0].data["L5"]
        dipole_times = dpl[0].times

        # Scale dipole to fit the spike raster plot
        raster_yrange = ax.get_yticks()
        raster_min = min(raster_yrange)
        raster_midpoint = round((raster_min / 2), 0)
        raster_quarterpoint = round((raster_min / 4), 0)

        # Scale down by .95 until the dipoles fit within the appropriate area
        while (
            max(max(l5_dipole), max(l2_dipole)) - min(min(l5_dipole), min(l2_dipole))
        ) > abs(raster_midpoint):
            l5_dipole = l5_dipole * 0.95
            l2_dipole = l2_dipole * 0.95

        # Shift the dipole positions to overlay the correct cell types
        l2_dipole = l2_dipole - abs(raster_midpoint) + abs(raster_quarterpoint)
        l5_dipole = l5_dipole - abs(raster_midpoint) - abs(raster_quarterpoint)

        # Draw the dipole plots
        (l2_line,) = ax.plot(
            dipole_times,
            l2_dipole,
            color="grey",
            linewidth=1,
            linestyle="--",
            label="L2 Dipole",
        )
        (l5_line,) = ax.plot(
            dipole_times,
            l5_dipole,
            color="grey",
            linewidth=1,
            linestyle="-",
            label="L5 Dipole",
        )

    # set legend
    handles = [e[0] for e in events]

    # Add labels for the spike events
    for handle in handles:
        handle.set_label(handle.get_label() + " Spikes")

    if overlay_dipoles:
        handles.append(l2_line)
        handles.append(l5_line)

    spike_legend = ax.legend(
        handles=handles,
        loc="lower left",
        framealpha=0.25,
    )
    if not show_legend:
        ax.get_legend().remove()
    else:
        ax.add_artist(spike_legend)

    # set axis labels
    ax.set_xlabel("Time (ms)")
    ax.set_ylabel("Cell ID")

    # hide y-axis ticks and tick labels
    ax.set_yticklabels([])
    ax.tick_params(axis="y", length=0)

    # add title
    if overlay_dipoles:
        ax.set_title("Raster Plot with Layer-Specific Dipole Overlays")
    else:
        ax.set_title("Raster Plot")

    if len(cell_response.times) > 0:
        ax.set_xlim(left=0, right=cell_response.times[-1])
    else:
        ax.set_xlim(left=0)
    ax.set_xlim(left=0)

    plt_show(show)
    return ax.get_figure()


def plot_cells(net, ax=None, show=True):
    """Plot the cells using Network.pos_dict.

    Parameters
    ----------
    net : instance of Network
        The Network object.
    ax : instance of matplotlib Axes3D | None
        An axis object from matplotlib. If None,
        a new figure is created.
    show : bool
        If True, show the figure.

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure handle.
    """
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

    elif not isinstance(ax, Axes3D):
        raise TypeError(
            f"Expected 'ax' to be an instance of Axes3D, but got {type(ax).__name__}"
        )

    colors = {
        "L5_pyramidal": "b",
        "L2_pyramidal": "c",
        "L5_basket": "r",
        "L2_basket": "m",
    }
    markers = {
        "L5_pyramidal": "^",
        "L2_pyramidal": "^",
        "L5_basket": "x",
        "L2_basket": "x",
    }

    for cell_type in net.cell_types:
        x = [pos[0] for pos in net.pos_dict[cell_type]]
        y = [pos[1] for pos in net.pos_dict[cell_type]]
        z = [pos[2] for pos in net.pos_dict[cell_type]]
        if cell_type in colors:
            color = colors[cell_type]
            marker = markers[cell_type]
            ax.scatter(x, y, z, c=color, s=50, marker=marker, label=cell_type)

    if net.rec_arrays:
        cols = plt.get_cmap("inferno", len(net.rec_arrays) + 2)
        for ii, (arr_name, arr) in enumerate(net.rec_arrays.items()):
            x = [p[0] for p in arr.positions]
            y = [p[1] for p in arr.positions]
            z = [p[2] for p in arr.positions]
            ax.scatter(x, y, z, color=cols(ii + 1), s=25, marker="o", label=arr_name)

    plt.legend(bbox_to_anchor=(-0.15, 1.025), loc="upper left")

    plt_show(show)
    return ax.get_figure()


def plot_tfr_morlet(
    dpl,
    freqs,
    *,
    n_cycles=7.0,
    tmin=None,
    tmax=None,
    layer="agg",
    decim=None,
    padding="zeros",
    ax=None,
    colormap="inferno",
    colorbar=True,
    colorbar_inside=False,
    show=True,
):
    """Plot Morlet time-frequency representation of dipole time course

    Parameters
    ----------
    dpl : instance of Dipole | list of Dipole instances
        The Dipole object. If a list of dipoles is given, the power is
        calculated separately for each trial, then averaged.
    freqs : array
        Frequency range of interest.
    n_cycles : float or array of float, default 7.0
        Number of cycles. Fixed number or one per frequency.
    tmin : float or None
        Start time of plot in milliseconds. If None, plot entire simulation.
    tmax : float or None
        End time of plot in milliseconds. If None, plot entire simulation.
    layer : str, default 'agg'
        The layer to plot. Can be one of 'agg', 'L2', and 'L5'
    decim : int or list of int or None (default)
        Optional (integer) factor by which to decimate the raw dipole traces.
        The SciPy function :func:`~scipy.signal.decimate` is used, which
        recommends values <13. To achieve higher decimation factors, a list of
        ints can be provided. These are applied successively.
    padding : str or None
        Optional padding of the dipole time course beyond the plotting limits.
        Possible values are: 'zeros' for padding with 0's (default), 'mirror'
        for mirror-image padding.
    ax : instance of matplotlib figure | None
        The matplotlib axis
    colormap : str
        The name of a matplotlib colormap, e.g., 'viridis'. Default: 'inferno'
    colorbar : bool
        If True (default), adjust figure to include colorbar.
    colorbar_inside: bool, default False
        Put the color inside the heatmap if True.
    show : bool
        If True, show the figure

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure handle.
    """
    import matplotlib.pyplot as plt
    from matplotlib.ticker import ScalarFormatter
    from .externals.mne import tfr_array_morlet
    from .dipole import Dipole

    if isinstance(dpl, Dipole):
        dpl = [dpl]

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)

    scale_applied = dpl[0].scale_applied
    sfreq = dpl[0].sfreq
    trial_power = []
    for dpl_trial in dpl:
        if dpl_trial.scale_applied != scale_applied:
            raise RuntimeError("All dipoles must be scaled equally!")
        if dpl_trial.sfreq != sfreq:
            raise RuntimeError("All dipoles must be sampled equally!")

        data, times = _get_plot_data_trange(
            dpl_trial.times, dpl_trial.data[layer], tmin, tmax
        )

        sfreq = dpl_trial.sfreq
        if decim is not None:
            data, times, sfreq = _decimate_plot_data(decim, data, times, sfreq=sfreq)

        if padding is not None:
            if not isinstance(padding, str):
                raise ValueError("padding must be a string (or None)")
            if padding == "zeros":
                data = np.r_[
                    np.zeros((len(data) - 1,)), data.ravel(), np.zeros((len(data) - 1,))
                ]
            elif padding == "mirror":
                data = np.r_[data[-1:0:-1], data, data[-2::-1]]

        # MNE expects an array of shape (n_trials, n_channels, n_times)
        data = data[None, None, :]
        power = tfr_array_morlet(
            data, sfreq=sfreq, freqs=freqs, n_cycles=n_cycles, output="power"
        )

        if padding is not None:
            # get the middle portion after padding
            power = power[:, :, :, times.shape[0] - 1 : 2 * times.shape[0] - 1]
        trial_power.append(power)

    power = np.mean(trial_power, axis=0)
    im = ax.pcolormesh(times, freqs, power[0, 0, ...], cmap=colormap, shading="auto")

    if freqs[0] > freqs[-1]:
        freqs = freqs[::-1]
        ax.invert_yaxis()

    ax.set_xlabel("Time (ms)")
    ax.set_ylabel("Frequency (Hz)")

    if colorbar:
        fig = ax.get_figure()
        xfmt = ScalarFormatter()
        xfmt.set_powerlimits((-2, 2))
        # default colorbar
        if colorbar_inside is False:
            cbar = fig.colorbar(im, ax=ax, format=xfmt, shrink=0.8, pad=0)
            cbar.ax.yaxis.set_ticks_position("left")
            cbar.ax.set_ylabel(
                r"Power ([nAm $\times$ {:.0f}]$^2$)".format(scale_applied),
                rotation=-90,
                va="bottom",
            )
        # put colorbar inside the heatmap.
        else:
            cbar_color = "white"
            cbar_fontsize = 6

            ax_pos = ax.get_position()
            ax_width = ax_pos.x1 - ax_pos.x0
            ax_height = ax_pos.y1 - ax_pos.y0
            cbar_L = ax_pos.x0 + 0.9 * ax_width
            cbar_B = ax_pos.y0 + 0.8 * ax_height
            cbar_W = ax_width * 0.04
            cbar_H = ax_height * 0.15

            cax = fig.add_axes([cbar_L, cbar_B, cbar_W, cbar_H])
            cbar = fig.colorbar(im, cax=cax, format=xfmt, shrink=0.8, pad=0)
            cbar.ax.yaxis.set_ticks_position("left")
            cbar.ax.yaxis.offsetText.set_fontsize(cbar_fontsize)

            cbar.ax.set_ylabel(
                r"Power ([nAm $\times$ {:.0f}]$^2$)".format(scale_applied),
                rotation=-90,
                va="bottom",
                fontsize=cbar_fontsize,
                color=cbar_color,
            )
            cbar.ax.tick_params(
                direction="in",
                labelsize=cbar_fontsize,
                labelcolor=cbar_color,
                colors=cbar_color,
            )
            plt.setp(cbar.ax.spines.values(), color=cbar_color)
            setattr(fig, f"_cbar-ax-{id(ax)}", cbar)

    plt_show(show)
    return ax.get_figure()


def plot_psd(
    dpl,
    *,
    fmin=0,
    fmax=None,
    tmin=None,
    tmax=None,
    layer="agg",
    color=None,
    label=None,
    ax=None,
    show=True,
):
    """Plot power spectral density (PSD) of dipole time course

    Applies `~scipy.signal.periodogram` from SciPy with ``window='hamming'``.
    Note that no spectral averaging is applied across time, as most
    ``hnn_core`` simulations are short-duration. However, passing a list of
    `Dipole` instances will plot their average (Hamming-windowed) power, which
    resembles the `Welch`-method applied over time.

    Parameters
    ----------
    dpl : instance of Dipole | list of Dipole instances
        The Dipole object.
    fmin : float
        Minimum frequency to plot (in Hz). Default: 0 Hz
    fmax : float
        Maximum frequency to plot (in Hz). Default: None (plot up to Nyquist)
    tmin : float or None
        Start time of data to include (in ms). If None, use entire simulation.
    tmax : float or None
        End time of data to include (in ms). If None, use entire simulation.
    layer : str, default 'agg'
        The layer to plot. Can be one of 'agg', 'L2', and 'L5'
    color : str or tuple or None
        The line color of PSD
    label : str or None
        Line label for PSD
    ax : instance of matplotlib figure | None
        The matplotlib axis.
    show : bool
        If True, show the figure

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure handle.
    """
    import matplotlib.pyplot as plt
    from scipy.signal import periodogram
    from .dipole import Dipole

    if ax is None:
        _, ax = plt.subplots(1, 1, constrained_layout=True)

    if isinstance(dpl, Dipole):
        dpl = [dpl]

    scale_applied = dpl[0].scale_applied
    sfreq = dpl[0].sfreq
    trial_power = []
    for dpl_trial in dpl:
        if dpl_trial.scale_applied != scale_applied:
            raise RuntimeError("All dipoles must be scaled equally!")
        if dpl_trial.sfreq != sfreq:
            raise RuntimeError("All dipoles must be sampled equally!")

        data, _ = _get_plot_data_trange(
            dpl_trial.times, dpl_trial.data[layer], tmin, tmax
        )

        freqs, Pxx = periodogram(data, sfreq, window="hamming", nfft=len(data))
        trial_power.append(Pxx)

    ax.plot(freqs, np.mean(np.array(Pxx, ndmin=2), axis=0), color=color, label=label)
    if label:
        ax.legend()
    if fmax is not None:
        ax.set_xlim((fmin, fmax))
    ax.ticklabel_format(axis="both", scilimits=(-2, 3))
    ax.set_xlabel("Frequency (Hz)")
    if scale_applied == 1:
        ylabel = "Power spectral density\n(nAm" + r"$^2 \ Hz^{-1}$)"
    else:
        ylabel = (
            "Power spectral density\n"
            + r"([nAm$\times$ {:.0f}]".format(scale_applied)
            + r"$^2 \ Hz^{-1}$)"
        )
    ax.set_ylabel(ylabel, multialignment="center")

    plt_show(show)
    return ax.get_figure()


def _linewidth_from_data_units(ax, linewidth):
    # see: https://stackoverflow.com/a/35501485
    fig = ax.get_figure()
    length = fig.bbox_inches.width * ax.get_position().width
    value_range = np.diff(ax.get_xlim())[0]
    length *= 72  # Convert length to points
    # Scale linewidth to value range
    return linewidth * (length / value_range)


def plot_cell_morphology(
    cell,
    ax,
    color=None,
    pos=(0, 0, 0),
    xlim=(-250, 150),
    ylim=(-100, 100),
    zlim=(-100, 1200),
    show=True,
):
    """Plot the cell morphology.

    Parameters
    ----------
    cell : instance of Cell
        The cell object
    ax : instance of Axes3D
        Matplotlib 3D axis
    show : bool
        If True, show the plot
    color : str | dict | None
        Color of cell. If str, entire cell plotted with
        color indicated by str. If dict, colors of individual sections
        can be specified. Must have a key for every section in cell as
        defined in the `Cell.sections` attribute.

        | Ex: ``{'apical_trunk': 'r', 'soma': 'b', ...}``
    pos : tuple of int or float | None
        Position of cell soma. Must be a tuple of 3 elements for the
        (x, y, z) position of the soma in 3D space. Default: (0, 0, 0)
    xlim : tuple of int | tuple of float
        x limits of plot window. Default (-250, 150)
    ylim : tuple of int | tuple of float
        y limits of plot window. Default (-100, 100)
    zlim : tuple of int | tuple of float
        z limits of plot window. Default (-100, 1200)
    show : bool
        If True, show the plot

    Returns
    -------
    axes : list of instance of Axes3D
        The matplotlib 3D axis handle.
    """
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D  # noqa

    if ax is None:
        plt.figure()
        ax = plt.axes(projection="3d")

    _validate_type(color, (str, dict, None), "color")

    if color is None:
        section_colors = {section: "b" for section in cell.sections.keys()}
    if isinstance(color, str):
        section_colors = {section: color for section in cell.sections.keys()}
    if isinstance(color, dict):
        section_colors = color

    _validate_type(pos, tuple, "pos")
    if isinstance(pos, tuple):
        if len(pos) != 3:
            raise ValueError("pos must be a tuple of 3 elements")
        for pos_idx in pos:
            _validate_type(pos_idx, (float, int), "pos[idx]")

    # Cell is in XZ plane
    ax.set_xlim((pos[0] + xlim[0], pos[0] + xlim[1]))
    ax.set_zlim((pos[1] + zlim[0], pos[1] + zlim[1]))
    ax.set_ylim((pos[2] + ylim[0], pos[2] + ylim[1]))

    for sec_name, section in cell.sections.items():
        linewidth = _linewidth_from_data_units(ax, section.diam)
        end_pts = section.end_pts
        dx = pos[0] - cell.sections["soma"].end_pts[0][0]
        dy = pos[1] - cell.sections["soma"].end_pts[0][1]
        dz = pos[2] - cell.sections["soma"].end_pts[0][2]
        xs, ys, zs = list(), list(), list()
        for pt in end_pts:
            xs.append(pt[0] + dx)
            ys.append(pt[1] + dz)
            zs.append(pt[2] + dy)
        ax.plot(xs, ys, zs, "-", linewidth=linewidth, color=section_colors[sec_name])
    ax.view_init(0, -90)
    ax.axis("off")

    plt.tight_layout()
    plt_show(show)
    return ax


def plot_connectivity_matrix(
    net, conn_idx, ax=None, show_weight=True, colorbar=True, colormap="Greys", show=True
):
    """Plot connectivity matrix with color bar for synaptic weights

    Parameters
    ----------
    net : Instance of Network object
        The Network object
    conn_idx : int
        Index of connection to be visualized
        from `net.connectivity`
    ax : instance of Axes3D
        Matplotlib 3D axis
    show_weight : bool
        If True, visualize connectivity weights as gradient.
        If False, all weights set to constant value.
    colormap : str
        The name of a matplotlib colormap. Default: 'Greys'
    colorbar : bool
        If True (default), adjust figure to include colorbar.
    show : bool
        If True, show the plot

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure handle.
    """
    import matplotlib.pyplot as plt
    from matplotlib.ticker import ScalarFormatter
    from .network import Network
    from .cell import _get_gaussian_connection

    _validate_type(net, Network, "net", "Network")
    _validate_type(conn_idx, int, "conn_idx", "int")
    _validate_type(show_weight, bool, "show_weight", "bool")
    if ax is None:
        _, ax = plt.subplots(1, 1)

    # Load objects for distance calculation
    conn = net.connectivity[conn_idx]
    nc_dict = conn["nc_dict"]
    src_type = conn["src_type"]
    target_type = conn["target_type"]
    src_type_pos = net.pos_dict[src_type]
    target_type_pos = net.pos_dict[target_type]

    src_range = np.array(net.gid_ranges[conn["src_type"]])
    target_range = np.array(net.gid_ranges[conn["target_type"]])
    connectivity_matrix = np.zeros((len(src_range), len(target_range)))

    for src_gid, target_src_pair in conn["gid_pairs"].items():
        src_idx = np.where(src_range == src_gid)[0][0]
        target_indeces = np.where(np.isin(target_range, target_src_pair))[0]
        for target_idx in target_indeces:
            src_pos = src_type_pos[src_idx]
            target_pos = target_type_pos[target_idx]

            # Identical calculation used in Cell.par_connect_from_src()
            if show_weight:
                weight, _ = _get_gaussian_connection(
                    src_pos, target_pos, nc_dict, inplane_distance=net._inplane_distance
                )
            else:
                weight = 1.0

            connectivity_matrix[src_idx, target_idx] = weight

    im = ax.imshow(connectivity_matrix, cmap=colormap, interpolation="none")

    ax.set_xlabel("Time (ms)")
    ax.set_ylabel("Frequency (Hz)")

    if colorbar:
        fig = ax.get_figure()
        xfmt = ScalarFormatter()
        xfmt.set_powerlimits((-2, 2))
        cbar = fig.colorbar(im, ax=ax, format=xfmt)
        cbar.ax.yaxis.set_ticks_position("right")
        cbar.ax.set_ylabel("Weight", rotation=-90, va="bottom")

    ax.set_xlabel(
        f"{conn['target_type']} target gids ({target_range[0]}-{target_range[-1]})"
    )
    ax.set_xticklabels(list())
    ax.set_ylabel(f"{conn['src_type']} source gids ({src_range[0]}-{src_range[-1]})")
    ax.set_yticklabels(list())
    ax.set_title(
        f"{conn['src_type']} -> {conn['target_type']} "
        f"({conn['loc']}, {conn['receptor']})"
    )

    plt.tight_layout()
    plt_show(show)
    return ax.get_figure()


def plot_drive_strength(
    net,
    show_weight=True,
    ax=None,
    colorbar=True,
    color_scale="linear",
    normalize=True,
    show=True,
):
    """Plot the relative strength of drives to cell types.

    Parameters
    ----------
    net : Network
        Instance of a Network object.
    show_weight : bool, default=True
        If True, visualize connectivity weights as gradient.
        If False, all weights set to constant value.
    ax : matplotlib.Axes, optional
        Matplotlib axes. If None, create a new figure.
    colorbar : bool, default=True
        If True (default), adjust figure to include colorbar.
    color_scale : {'linear', 'log'}, default='linear'
        Color scaling for drive strength.
    normalize : bool, default=True
        If True, normalize the strength values to be between 0 and 1.
    show : bool, default=True
        If True, show the plot immediately.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure handle.
    """
    import numpy as np
    import matplotlib.pyplot as plt
    from .network import Network
    from .cell import _get_gaussian_connection

    _validate_type(net, Network, "net", "Network")
    _validate_type(show_weight, bool, "show_weight", "bool")

    # Get drive and cell type information
    drive_names = list(net.external_drives.keys())
    cell_types = list(net.cell_types.keys())

    strength_matrix = np.zeros((len(drive_names), len(cell_types)))

    for conn in net.connectivity:
        src_type = conn["src_type"]
        target_type = conn["target_type"]
        if src_type in drive_names and target_type in cell_types:
            nc_dict = conn["nc_dict"]
            drive_idx = drive_names.index(src_type)
            cell_idx = cell_types.index(target_type)
            src_type_pos = net.pos_dict[src_type]
            target_type_pos = net.pos_dict[target_type]

            src_range = np.array(net.gid_ranges[src_type])
            target_range = np.array(net.gid_ranges[target_type])

            total_weight = 0
            for src_gid, target_src_pair in conn["gid_pairs"].items():
                src_idx = np.where(src_range == src_gid)[0][0]
                target_indeces = np.where(np.isin(target_range, target_src_pair))[0]
                for target_idx in target_indeces:
                    src_pos = src_type_pos[src_idx]
                    target_pos = target_type_pos[target_idx]

                    if show_weight:
                        weight = _get_gaussian_connection(
                            src_pos,
                            target_pos,
                            nc_dict,
                            inplane_distance=net._inplane_distance,
                        )
                    else:
                        weight = 1.0

                    total_weight += sum(weight)
                strength_matrix[drive_idx, cell_idx] = total_weight

    if normalize:
        min_val = np.min(strength_matrix)
        max_val = np.max(strength_matrix)
        if max_val > min_val:
            strength_matrix = (strength_matrix - min_val) / (max_val - min_val)

    if color_scale == "log":
        strength_matrix = np.log1p(strength_matrix)

    if ax is None:
        fig, ax = plt.subplots(figsize=(12, 6))
    else:
        fig = ax.figure

    cax = ax.imshow(
        strength_matrix, cmap="viridis", aspect="auto", interpolation="nearest"
    )

    if colorbar:
        cbar = fig.colorbar(cax, ax=ax)
        cbar.ax.yaxis.set_ticks_position("right")

    ax.set_xticks(np.arange(len(cell_types)))
    ax.set_yticks(np.arange(len(drive_names)))
    ax.set_xticklabels(cell_types)
    ax.set_yticklabels(drive_names)
    ax.set_xlabel("Cell Types")
    ax.set_ylabel("Drive Names")
    ax.set_title("External Drive Strengths")

    # For addition of units information
    unit_text = (
        "Units: µS (microsiemens)"
        if not normalize
        else "Units: Relative Strength (0-1)"
    )
    ax.text(
        0.98,
        0.02,
        unit_text,
        transform=ax.transAxes,
        fontsize=12,
        verticalalignment="bottom",
        horizontalalignment="right",
        bbox=dict(facecolor="white", alpha=0.6, edgecolor="black"),
    )

    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

    for i in range(len(drive_names)):
        for j in range(len(cell_types)):
            value = strength_matrix[i, j]
            color = "black" if value > strength_matrix.max() * 0.7 else "white"
            ax.text(
                j,
                i,
                f"{value:.2f}",
                ha="center",
                va="center",
                color=color,
                fontsize=12,
                weight="bold",
            )

    plt.tight_layout()
    if show:
        plt.show()
    return fig


def _update_target_plot(
    ax,
    conn,
    src_gid,
    src_type_pos,
    target_type_pos,
    src_range,
    target_range,
    nc_dict,
    colormap,
    inplane_distance,
):
    from .cell import _get_gaussian_connection

    # Extract indices to get position in network
    # Index in gid range aligns with net.pos_dict
    target_src_pair = conn["gid_pairs"][src_gid]
    target_indeces = np.where(np.isin(target_range, target_src_pair))[0]

    src_idx = np.where(src_range == src_gid)[0][0]
    src_pos = src_type_pos[src_idx]

    # Aggregate positions and weight of each connected target
    weights, target_x_pos, target_y_pos = list(), list(), list()
    for target_idx in target_indeces:
        target_pos = target_type_pos[target_idx]
        target_x_pos.append(target_pos[0])
        target_y_pos.append(target_pos[1])
        weight, _ = _get_gaussian_connection(
            src_pos, target_pos, nc_dict, inplane_distance
        )
        weights.append(weight)

    ax.clear()
    im = ax.scatter(target_x_pos, target_y_pos, c=weights, s=50, cmap=colormap)
    x_pos = target_type_pos[:, 0]
    y_pos = target_type_pos[:, 1]
    ax.scatter(x_pos, y_pos, color="k", marker="x", zorder=-1, s=20)
    ax.scatter(src_pos[0], src_pos[1], marker="s", color="red", s=150)
    ax.set_ylabel("Y Position")
    ax.set_xlabel("X Position")
    return im


def plot_cell_connectivity(
    net, conn_idx, src_gid=None, axes=None, colorbar=True, colormap="viridis", show=True
):
    """Plot synaptic weight of connections.

    This is an interactive plot with source cells shown in the left
    subplot and connectivity from a source cell to all the target cells
    in the right subplot. Click on the cells in the left subplot to
    explore how the connectivity pattern changes for different source cells.

    Parameters
    ----------
    net : Instance of Network object
        The Network object
    conn_idx : int
        Index of connection to be visualized from net.connectivity
    src_gid : int | None
        The cell ID of the source cell. It must be an element of
        net.connectivity[conn_idx]['gid_pairs'].keys()
        If None, the first cell from the list of valid src_gids is selected.
    axes : instance of Axes3D
        Matplotlib 3D axis
    colormap : str
        The name of a matplotlib colormap. Default: 'viridis'
    colorbar : bool
        If True (default), adjust figure to include colorbar.
    show : bool
        If True, show the plot

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure handle.

    Notes
    -----
    Target cells will be determined by the connections in
    net.connectivity[conn_idx].
    If the target cell is not connected to the source cell,
    it will appear as a smaller black cross.
    Source cell is plotted as a red square. Source cell will not be plotted if
    the connection corresponds to a drive, ex: poisson, bursty, etc.

    """
    import matplotlib.pyplot as plt
    from .network import Network
    from matplotlib.ticker import ScalarFormatter

    _validate_type(net, Network, "net", "Network")
    _validate_type(conn_idx, int, "conn_idx", "int")

    # Load objects for distance calculation
    conn = net.connectivity[conn_idx]
    nc_dict = conn["nc_dict"]
    src_type = conn["src_type"]
    target_type = conn["target_type"]
    src_type_pos = np.array(net.pos_dict[src_type])
    target_type_pos = np.array(net.pos_dict[target_type])
    src_range = np.array(net.gid_ranges[conn["src_type"]])

    valid_src_gids = list(net.connectivity[conn_idx]["gid_pairs"].keys())
    src_pos_valid = src_type_pos[np.isin(src_range, valid_src_gids)]

    if src_gid is None:
        src_gid = valid_src_gids[0]
    _validate_type(src_gid, int, "src_gid", "int")

    if src_gid not in valid_src_gids:
        raise ValueError(
            f"src_gid {src_gid} not a valid cell ID for this "
            f"connection. Please select one of {valid_src_gids}"
        )

    target_range = np.array(net.gid_ranges[conn["target_type"]])

    if axes is None:
        if src_type in net.cell_types:
            fig, axes = plt.subplots(1, 2, sharex=True, sharey=True)
        else:
            fig, axes = plt.subplots(1, 1, sharex=True, sharey=True)
            axes = [axes]

    if len(axes) == 2:
        ax_src, ax = axes
    else:
        ax = axes[0]

    im = _update_target_plot(
        ax,
        conn,
        src_gid,
        src_type_pos,
        target_type_pos,
        src_range,
        target_range,
        nc_dict,
        colormap,
        net._inplane_distance,
    )

    x_src = src_type_pos[:, 0]
    y_src = src_type_pos[:, 1]
    x_src_valid = src_pos_valid[:, 0]
    y_src_valid = src_pos_valid[:, 1]
    if src_type in net.cell_types:
        ax_src.scatter(x_src, y_src, marker="s", color="red", s=50, alpha=0.2)
        ax_src.scatter(x_src_valid, y_src_valid, marker="s", color="red", s=50)

    plt.suptitle(
        f"{conn['src_type']}-> {conn['target_type']}"
        f" ({conn['loc']}, {conn['receptor']})"
    )

    def _onclick(event):
        if event.inaxes in [ax] or event.inaxes is None:
            return

        dist = np.linalg.norm(
            src_type_pos[:, :2] - np.array([event.xdata, event.ydata]), axis=1
        )
        src_idx = np.argmin(dist)

        src_gid = src_range[src_idx]
        if src_gid not in valid_src_gids:
            return
        _update_target_plot(
            ax,
            conn,
            src_gid,
            src_type_pos,
            target_type_pos,
            src_range,
            target_range,
            nc_dict,
            colormap,
            net._inplane_distance,
        )

        fig.canvas.draw()

    if colorbar:
        fig = ax.get_figure()
        xfmt = ScalarFormatter()
        xfmt.set_powerlimits((-2, 2))
        cbar = fig.colorbar(im, ax=ax, format=xfmt)
        cbar.ax.yaxis.set_ticks_position("right")
        cbar.ax.set_ylabel("Weight", rotation=-90, va="bottom")

    plt.tight_layout()

    fig.canvas.mpl_connect("button_press_event", _onclick)

    plt_show(show)
    return ax.get_figure()


def plot_laminar_csd(
    times,
    data,
    contact_labels,
    ax=None,
    colorbar=True,
    vmin=None,
    vmax=None,
    sink="b",
    interpolation="spline",
    show=True,
):
    """Plot laminar current source density (CSD) estimation from LFP array.

    Parameters
    ----------
    times : Numpy array, shape (n_times,)
        Sampling times (in ms).
    data : array-like, shape (n_channels, n_times)
        CSD data, channels x time.
    ax : instance of matplotlib figure | None
        The matplotlib axis.
    colorbar : bool
        If True (default), adjust figure to include colorbar.
    contact_labels : list
        Labels associated with the contacts to plot. Passed as-is to
        :func:`~matplotlib.axes.Axes.set_yticklabels`.
    vmin: float, optional
        lower bound of the color axis.
        Will be set automatically of None.
    vmax: float, optional
        upper bound of the color axis.
        Will be set automatically of None.
    sink : str
        If set to 'blue' or 'b', plots sinks in blue and sources in red,
        if set to 'red' or 'r', sinks plotted in red and sources blue.
    interpolation : str | None
        If 'spline', will smoothen the CSD using spline method,
        if None, no smoothing will be applied.

    show : bool
        If True, show the plot.

    Returns
    -------
    fig : instance of matplotlib Figure
        The matplotlib figure handle.
    """
    import matplotlib.pyplot as plt
    from scipy.interpolate import RectBivariateSpline

    if ax is None:
        _, ax = plt.subplots(1, 1, constrained_layout=True)

    if sink[0].lower() == "b":
        cmap = "jet"
    elif sink[0].lower() == "r":
        cmap = "jet_r"
    elif sink[0].lower() != "b" or sink[0].lower() != "r":
        raise RuntimeError(
            'Please use sink = "b" or sink = "r".'
            ' Only colormap "jet" is supported for CSD.'
        )

    if interpolation == "spline":
        # create interpolation function
        interp_data = RectBivariateSpline(times, contact_labels, data.T)
        # increase number of contacts
        new_depths = np.linspace(
            contact_labels[0],
            contact_labels[-1],
            contact_labels[-1] - contact_labels[0],
        )
        # interpolate
        data = interp_data(times, new_depths).T
    elif interpolation is None:
        data = data
        new_depths = contact_labels

    # if vmin and vmax are both None, set colormap such that green = zero
    if vmin is None and vmax is None:
        vmin = -np.max(np.abs(data))
        vmax = np.max(np.abs(data))

    im = ax.pcolormesh(
        times, new_depths, data, cmap=cmap, shading="auto", vmin=vmin, vmax=vmax
    )
    ax.set_xlabel("time (s)")
    ax.set_ylabel("electrode depth")
    if colorbar:
        color_axis = ax.inset_axes([1.05, 0, 0.02, 1], transform=ax.transAxes)
        plt.colorbar(im, ax=ax, cax=color_axis).set_label(r"$CSD (uV/um^{2})$")

    plt.tight_layout()
    plt_show(show)

    return ax.get_figure()


class NetworkPlotter:
    """Helper class to visualize full morphology of HNN model.

    Parameters
    ----------
    net : Instance of Network object
        The Network object
    ax : instance of matplotlib Axes3D | None
        An axis object from matplotlib. If None,
        a new figure is created.
    vmin : int | float
        Lower limit of colormap for plotting voltage
        Default: -100 mV
    vmax : int | float
        Upper limit of colormap for plotting voltage
        Default: 50 mV
    bg_color : str
        Background color of ax. Default: 'black'
    colorbar : bool
        If True (default), adjust figure to include colorbar.
    voltage_colormap : str
        Colormap used for plotting voltages
        Default: 'viridis'
    elev : int | float
        Elevation 3D plot viewpoint, default: 10
    azim : int | float
        Azimuth of 3D plot view point, default: 20
    xlim : tuple of int | tuple of float
        x limits of plot window. Default (-200, 3100)
    ylim : tuple of int | tuple of float
        y limits of plot window. Default (-200, 3100)
    zlim : tuple of int | tuple of float
        z limits of plot window. Default (-300, 2200)
    trial_idx : int
        Index of simulation trial plotted. Default: 0
    time_idx : int
        Index of time point plotted. Default: 0
    """

    def __init__(
        self,
        net,
        ax=None,
        vmin=-100,
        vmax=50,
        bg_color="black",
        colorbar=True,
        voltage_colormap="viridis",
        elev=10,
        azim=-500,
        xlim=(-200, 3100),
        ylim=(-200, 3100),
        zlim=(-300, 2200),
        trial_idx=0,
        time_idx=0,
    ):
        from matplotlib import colormaps

        self._validate_parameters(
            vmin,
            vmax,
            bg_color,
            voltage_colormap,
            colorbar,
            elev,
            azim,
            xlim,
            ylim,
            zlim,
            trial_idx,
            time_idx,
        )

        # Set init arguments
        self.net = net
        self.ax = ax
        self._vmin = vmin
        self._vmax = vmax
        self._bg_color = bg_color
        self._colorbar = colorbar
        self._voltage_colormap = voltage_colormap
        self._colormaps = colormaps
        self._xlim = xlim
        self._ylim = ylim
        self._zlim = zlim
        self._elev = elev
        self._azim = azim
        self._trial_idx = trial_idx
        self._time_idx = time_idx

        # Check if Network object is simulated
        self.times, self._vsec_recorded = self._check_network_simulation()

        # Initialize plots and colormap
        self.fig = None
        self._colormap = colormaps[voltage_colormap]
        self.vsec_array = self._get_voltages()
        self.color_array = self._colormap(self.vsec_array)

        self._initialize_plots()
        if self._colorbar:
            self._update_colorbar()
        else:
            self._cbar = None

    def _validate_parameters(
        self,
        vmin,
        vmax,
        bg_color,
        voltage_colormap,
        colorbar,
        elev,
        azim,
        xlim,
        ylim,
        zlim,
        trial_idx,
        time_idx,
    ):
        _validate_type(vmin, (int, float), "vmin")
        _validate_type(vmax, (int, float), "vmax")
        _validate_type(bg_color, str, "bg_color")
        _validate_type(voltage_colormap, str, "voltage_colormap")
        _validate_type(colorbar, bool, "colorbar")
        _validate_type(xlim, tuple, "xlim")
        _validate_type(ylim, tuple, "ylim")
        _validate_type(zlim, tuple, "zlim")
        _validate_type(elev, (int, float), "elev")
        _validate_type(azim, (int, float), "azim")
        _validate_type(trial_idx, int, "trial_idx")
        _validate_type(time_idx, int, "time_idx")

    def _check_network_simulation(self):
        times = None
        vsec_recorded = False
        # Check if network simulated
        if self.net.cell_response is not None:
            times = self.net.cell_response.times

            # Check if voltage recorded
            if self.net._params["record_vsec"] == "all":
                vsec_recorded = True
        return times, vsec_recorded

    def _initialize_plots(self):
        import matplotlib.pyplot as plt

        # Create figure
        if self.ax is None:
            self.fig = plt.figure()
            self.ax = self.fig.add_subplot(projection="3d")
            self.ax.set_facecolor(self._bg_color)

        self._init_network_plot()
        self._update_axes()

    def _get_voltages(self):
        vsec_list = list()
        for cell_type in self.net.cell_types:
            gid_range = self.net.gid_ranges[cell_type]
            for gid in gid_range:
                cell = self.net.cell_types[cell_type]["cell_object"]
                for sec_name in cell.sections.keys():
                    if self._vsec_recorded is True:
                        vsec = np.array(
                            self.net.cell_response.vsec[self.trial_idx][gid][sec_name]
                        )
                        vsec_list.append(vsec)
                    else:  # Populate with zeros if no voltage recording
                        vsec_list.append([0.0])

        vsec_array = np.vstack(vsec_list)
        vsec_array = (vsec_array - self.vmin) / (self.vmax - self.vmin)
        return vsec_array

    def _update_section_voltages(self, t_idx):
        if not self._vsec_recorded:
            raise RuntimeError(
                "Network must be simulated with"
                "`simulate_dipole(record_vsec='all')` before"
                "plotting voltages."
            )
        color_list = self.color_array[:, t_idx]
        for line, color in zip(self.ax.lines, color_list):
            line.set_color(color)

    def _init_network_plot(self):
        for cell_type in self.net.cell_types:
            gid_range = self.net.gid_ranges[cell_type]
            for gid_idx, gid in enumerate(gid_range):
                cell = self.net.cell_types[cell_type]

                pos = self.net.pos_dict[cell_type][gid_idx]
                pos = (float(pos[0]), float(pos[2]), float(pos[1]))

                cell["cell_object"].plot_morphology(
                    ax=self.ax,
                    show=False,
                    pos=pos,
                    xlim=self.xlim,
                    ylim=self.ylim,
                    zlim=self.zlim,
                )

    def _update_axes(self):
        self.ax.set_xlim(self._xlim)
        self.ax.set_ylim(self._ylim)
        self.ax.set_zlim(self._zlim)

        self.ax.view_init(self._elev, self._azim)

    def _update_colorbar(self):
        import matplotlib.pyplot as plt
        import matplotlib.colors as mc

        fig = self.ax.get_figure()
        sm = plt.cm.ScalarMappable(
            cmap=self.voltage_colormap,
            norm=mc.Normalize(vmin=self.vmin, vmax=self.vmax),
        )
        self._cbar = fig.colorbar(sm, ax=self.ax)

    def export_movie(
        self,
        fname,
        fps=30,
        dpi=300,
        decim=10,
        interval=30,
        frame_start=0,
        frame_stop=None,
        writer="pillow",
    ):
        """Export movie of network activity

        Parameters
        ----------
        fname : str
            Filename of exported movie
        fps : int
            Frames per second, default: 30
        dpi : int
            Dots per inch, default: 300
        decim : int
            Decimation factor for frames, default: 10
        interval : int
            Delay between frames, default: 30
        frame_start : int
            Index of first frame, default: 0
        frame_stop : int | None
            Index of last frame, default: None
            If None, entire simulation is animated
        writer : str
            Movie writer, default: 'pillow'.
            Alternative movie writers can be found at
            https://matplotlib.org/stable/api/animation_api.html
        """
        import matplotlib.animation as animation

        if not self._vsec_recorded:
            raise RuntimeError(
                "Network must be simulated with"
                "`simulate_dipole(record_vsec='all')` before"
                "plotting voltages."
            )
        if frame_stop is None:
            frame_stop = len(self.times) - 1

        frames = np.arange(frame_start, frame_stop, decim)
        ani = animation.FuncAnimation(
            self.fig, self._set_time_idx, frames, interval=interval
        )

        writer = animation.writers[writer](fps=fps)
        ani.save(fname, writer=writer, dpi=dpi)
        return ani

    # Axis limits
    @property
    def xlim(self):
        return self._xlim

    @xlim.setter
    def xlim(self, xlim):
        _validate_type(xlim, tuple, "xlim")
        self._xlim = xlim
        self.ax.set_xlim(self._xlim)

    @property
    def ylim(self):
        return self._ylim

    @ylim.setter
    def ylim(self, ylim):
        _validate_type(ylim, tuple, "ylim")
        self._ylim = ylim
        self.ax.set_ylim(self._ylim)

    @property
    def zlim(self):
        return self._zlim

    @zlim.setter
    def zlim(self, zlim):
        _validate_type(zlim, tuple, "zlim")
        self._zlim = zlim
        self.ax.set_zlim(self._zlim)

    # Elevation and azimuth of 3D viewpoint
    @property
    def elev(self):
        return self._elev

    @elev.setter
    def elev(self, elev):
        _validate_type(elev, (int, float), "elev")
        self._elev = elev
        self.ax.view_init(self._elev, self._azim)

    @property
    def azim(self):
        return self._azim

    @azim.setter
    def azim(self, azim):
        _validate_type(azim, (int, float), "azim")
        self._azim = azim
        self.ax.view_init(self._elev, self._azim)

    # Minimum and maximum voltages
    @property
    def vmin(self):
        return self._vmin

    @vmin.setter
    def vmin(self, vmin):
        _validate_type(vmin, (int, float), "vmin")
        self._vmin = vmin
        self.vsec_array = self._get_voltages()
        self.color_array = self._colormap(self.vsec_array)
        if self._colorbar:
            self._cbar.remove()
            self._update_colorbar()

    @property
    def vmax(self):
        return self._vmax

    @vmax.setter
    def vmax(self, vmax):
        _validate_type(vmax, (int, float), "vmax")
        self._vmax = vmax
        self.vsec_array = self._get_voltages()
        self.color_array = self._colormap(self.vsec_array)
        if self._colorbar:
            self._cbar.remove()
            self._update_colorbar()

    # Time and trial indices
    @property
    def trial_idx(self):
        return self._trial_idx

    @trial_idx.setter
    def trial_idx(self, trial_idx):
        _validate_type(trial_idx, int, "trial_idx")
        if not self._vsec_recorded:
            raise RuntimeError(
                "Network must be simulated with"
                "`simulate_dipole(record_vsec='all')` before"
                "setting `trial_idx`."
            )
        self._trial_idx = trial_idx
        self.vsec_array = self._get_voltages()
        self.color_array = self._colormap(self.vsec_array)
        self._update_section_voltages(self._time_idx)

    @property
    def time_idx(self):
        return self._time_idx

    @time_idx.setter
    def time_idx(self, time_idx):
        _validate_type(time_idx, (int, np.integer), "time_idx")
        if not self._vsec_recorded:
            raise RuntimeError(
                "Network must be simulated with"
                "`simulate_dipole(record_vsec='all')` before"
                "setting `time_idx`."
            )
        self._time_idx = time_idx
        self._update_section_voltages(self._time_idx)

    # Callable update function for making animations
    def _set_time_idx(self, time_idx):
        self.time_idx = time_idx

    # Background color and voltage colormaps
    @property
    def bg_color(self):
        return self._bg_color

    @bg_color.setter
    def bg_color(self, bg_color):
        self._bg_color = bg_color
        self.ax.set_facecolor(self._bg_color)

    @property
    def voltage_colormap(self):
        return self._voltage_colormap

    @voltage_colormap.setter
    def voltage_colormap(self, voltage_colormap):
        self._voltage_colormap = voltage_colormap
        self._colormap = self._colormaps[self._voltage_colormap]
        self.color_array = self._colormap(self.vsec_array)
        if self._colorbar:
            self._cbar.remove()
            self._update_colorbar()

    @property
    def colorbar(self):
        return self._colorbar

    @colorbar.setter
    def colorbar(self, colorbar):
        _validate_type(colorbar, bool, "colorbar")
        self._colorbar = colorbar
        if self._colorbar:
            # Remove old colorbar if already exists
            if self._cbar is not None:
                self._cbar.remove()
            self._update_colorbar()
        else:
            self._cbar.remove()
            self._cbar = None



"""Utility functions."""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>
#          Christopher Bailey <cjb@cfin.au.dk>

import numpy as np
from .externals.mne import _validate_type


def _hammfilt(x, winsz):
    """Convolve with a hamming window."""
    assert len(x) > winsz
    win = np.hamming(winsz)
    win /= sum(win)
    return np.convolve(x, win, "same")


def _replace_dict_identifier(input_dict, old_identifier, new_identifier):
    """Recursively replace keys and values in a dict that match an identifier.

    This takes an `old_identifier` (anything that can compared to any keys or
    values with an equivalence relation), recurses through a dictionary
    (including any nested lists or dicts), and replaces all instances of
    `old_identifier` with `new_identifier`, regardless of if they are keys or
    values.

    Parameters
    ----------
    input_dict : dict
        The dictionary to update.
    old_identifier : str
        The key or value to replace.
    new_identifier : str
        The new key or value to replace with.

    Returns
    -------
    updated_dict : dict
        The updated dictionary with keys and values replaced.
    """
    if isinstance(input_dict, dict):
        updated_dict = dict()
        for key, value in input_dict.items():
            new_key_name = new_identifier if key == old_identifier else key
            updated_dict[new_key_name] = _replace_dict_identifier(
                value, old_identifier, new_identifier
            )
        return updated_dict
    elif isinstance(input_dict, list):
        return [
            _replace_dict_identifier(element, old_identifier, new_identifier)
            for element in input_dict
        ]
    else:
        return new_identifier if input_dict == old_identifier else input_dict


# Savitzky-Golay filtering, lifted and adapted from mne-python (0.22)
def _savgol_filter(data, h_freq, sfreq):
    """Filter the data using Savitzky-Golay polynomial method.

    Parameters
    ----------
    data : array-like
        The data to filter (1D)
    h_freq : float
        Approximate high cutoff frequency in Hz. Note that this
        is not an exact cutoff, since Savitzky-Golay filtering
        is done using polynomial fits
        instead of FIR/IIR filtering. This parameter is thus used to
        determine the length of the window over which a 5th-order
        polynomial smoothing is applied.
    sfreq : float
        The sampling frequency (in Hz)

    Returns
    -------
    filt_data : array-like
        The filtered data
    """  # noqa: E501
    from scipy.signal import savgol_filter

    _validate_type(sfreq, (float, int), "sfreq")
    assert sfreq > 0.0
    _validate_type(h_freq, (float, int), "h_freq")
    assert h_freq > 0.0

    h_freq = float(h_freq)
    if h_freq >= sfreq / 2.0:
        raise ValueError("h_freq must be less than half the sample rate")

    # savitzky-golay filtering
    window_length = (int(np.round(sfreq / h_freq)) // 2) * 2 + 1
    # loop over 'agg', 'L2', and 'L5'
    filt_data = savgol_filter(data, axis=-1, polyorder=5, window_length=window_length)
    return filt_data


def smooth_waveform(data, window_len, sfreq):
    """Smooth an arbitrary waveform using Hamming-windowed convolution

    Parameters
    ----------
    data : list | np.ndarray
        The data to filter
    window_len : float
        The length (in ms) of a `~numpy.hamming` window to convolve the
        data with.
    sfreq : float
        The data sampling rate.

    Returns
    -------
    data_filt : np.ndarray
        The filtered data
    """
    if (isinstance(data, np.ndarray) and data.ndim > 1) or (
        isinstance(data, list) and isinstance(data[0], list)
    ):
        raise RuntimeError("smoothing currently only supported for 1D-arrays")

    if not isinstance(window_len, (float, int)) or window_len < 0:
        raise ValueError("Window length must be a non-negative number")
    elif 0 < window_len < 1:
        raise ValueError("Window length less than 1 ms is not supported")

    _validate_type(sfreq, (float, int), "sfreq")
    assert sfreq > 0.0
    # convolutional filter length is given in samples
    winsz = np.round(1e-3 * window_len * sfreq)
    if winsz > len(data):
        raise ValueError(
            f"Window length too long: {winsz} samples; data length is "
            f"{len(data)} samples"
        )

    return _hammfilt(data, winsz)



# Authors: Nick Tolley <nicholas_tolley@brown.edu>
#          Christopher Bailey <cjb@cfin.au.dk>

from copy import deepcopy
import os.path as op
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest

import hnn_core
from hnn_core import read_params, jones_2009_model, simulate_dipole
from hnn_core.extracellular import (
    ExtracellularArray,
    calculate_csd2d,
    _get_laminar_z_coords,
)
from hnn_core.parallel_backends import requires_mpi4py, requires_psutil

import matplotlib.pyplot as plt


hnn_core_root = op.dirname(hnn_core.__file__)
params_fname = op.join(hnn_core_root, "param", "default.json")
params = read_params(params_fname)


def test_extracellular_api():
    """Test extracellular recording API."""
    net = jones_2009_model(deepcopy(params), add_drives_from_params=True)

    # Test LFP electrodes
    electrode_pos = (1, 2, 3)
    net.add_electrode_array("el1", electrode_pos)
    electrode_pos = [(1, 2, 3), (-1, -2, -3)]
    net.add_electrode_array("arr1", electrode_pos)
    assert len(net.rec_arrays) == 2
    assert len(net.rec_arrays["arr1"].positions) == 2

    # Test other not NotImplemented for ExtracellularArray Class
    assert (net.rec_arrays["arr1"] == "extArr") is False

    # ensure unique names
    pytest.raises(ValueError, net.add_electrode_array, "arr1", [(6, 6, 800)])
    # all remaining input arguments checked by ExtracellularArray

    rec_arr = ExtracellularArray(electrode_pos)

    # Added second string in the match pattern due to changes in python >=3.11
    # AttributeError message changed to "property X of object Y has no setter"
    with pytest.raises(AttributeError, match="has no setter|can't set attribute"):
        rec_arr.times = [1, 2, 3]
    with pytest.raises(AttributeError, match="has no setter|can't set attribute"):
        rec_arr.voltages = [1, 2, 3]
    with pytest.raises(TypeError, match="trial index must be int"):
        _ = rec_arr["0"]
    with pytest.raises(IndexError, match="the data contain"):
        _ = rec_arr[42]

    # positions are 3-tuples
    bad_positions = [[(1, 2), (1, 2, 3)], [42, (1, 2, 3)]]
    for bogus_pos in bad_positions:
        pytest.raises((ValueError, TypeError), ExtracellularArray, bogus_pos)

    good_positions = [(1, 2, 3), (100, 200, 300)]
    for cond in ["0.3", [0.3], -1]:  # conductivity is positive float
        pytest.raises(
            (TypeError, AssertionError),
            ExtracellularArray,
            good_positions,
            conductivity=cond,
        )
    for meth in ["foo", 0.3]:  # method is 'psa' or 'lsa' (or None for test)
        pytest.raises(
            (TypeError, AssertionError, ValueError),
            ExtracellularArray,
            good_positions,
            method=meth,
        )
    for mind in ["foo", -1, None]:  # minimum distance to segment boundary
        pytest.raises(
            (TypeError, AssertionError),
            ExtracellularArray,
            good_positions,
            min_distance=mind,
        )

    pytest.raises(
        ValueError,
        ExtracellularArray,  # more chans than voltages
        good_positions,
        times=[1],
        voltages=[[[42]]],
    )
    pytest.raises(
        ValueError,
        ExtracellularArray,  # less times than voltages
        good_positions,
        times=[1],
        voltages=[[[42, 42], [84, 84]]],
    )

    rec_arr = ExtracellularArray(
        good_positions,
        times=[0, 0.1, 0.21, 0.3],  # uneven sampling
        voltages=[[[0, 0, 0, 0], [0, 0, 0, 0]]],
    )
    with pytest.raises(RuntimeError, match="Extracellular sampling times"):
        _ = rec_arr.sfreq
    rec_arr._reset()
    assert len(rec_arr.times) == len(rec_arr.voltages) == 0
    assert rec_arr.sfreq is None
    rec_arr = ExtracellularArray(good_positions, times=[0], voltages=[[[0], [0]]])
    with pytest.raises(RuntimeError, match="Sampling rate is not defined"):
        _ = rec_arr.sfreq

    # test colinearity and equal spacing between electrode contacts for laminar
    # profiling (e.g., for platting laminar LFP or CSD)
    electrode_pos = [(1, 2, 1000), (2, 3, 3000), (3, 4, 5000), (4, 5, 7000)]
    z_coords, z_delta = _get_laminar_z_coords(electrode_pos)
    assert np.array_equal(z_coords, [1000, 3000, 5000, 7000])
    assert z_delta == 2000
    with pytest.raises(
        ValueError, match="Electrode array positions must contain more than 1 contact"
    ):
        _, _ = _get_laminar_z_coords([(1, 2, 3)])
    with pytest.raises(
        ValueError, match="Make sure the electrode positions are equispaced, colinear"
    ):
        _, _ = _get_laminar_z_coords([(1, 1, 3), (1, 1, 4), (1, 1, 3.5)])


def test_transmembrane_currents():
    """Test that net transmembrane current is zero at all times."""
    params.update(
        {
            "N_pyr_x": 3,
            "N_pyr_y": 3,
            "t_evprox_1": 5,
            "t_evdist_1": 10,
            "t_evprox_2": 20,
            "N_trials": 1,
        }
    )
    net = jones_2009_model(params, add_drives_from_params=True)
    electrode_pos = (0, 0, 0)  # irrelevant where electrode is
    # all transfer resistances set to unity
    net.add_electrode_array("net_Im", electrode_pos, method=None)
    _ = simulate_dipole(net, tstop=40.0)
    assert_allclose(net.rec_arrays["net_Im"].voltages, 0, rtol=1e-10, atol=1e-10)


def test_transfer_resistance():
    """Test transfer resistances calculated correctly"""
    from neuron import h
    from hnn_core.extracellular import _transfer_resistance

    sec = h.Section(name="dend")
    h.pt3dclear(sec=sec)
    h.pt3dadd(0, 0, 0, 1, sec=sec)
    h.pt3dadd(0, 1, 0, 1, sec=sec)  # section oriented along y-axis
    sec.L = 300
    sec.diam = 8
    sec.nseg = 5
    # NB segment lengths aren't equal! First/last segment center point is
    # closer to respective end point than to next/previous segment!
    seg_ctr_pts = [0]
    seg_ctr_pts.extend([seg.x * sec.L for seg in sec])
    seg_ctr_pts.append(sec.L)
    seg_lens = np.diff(seg_ctr_pts)
    first_len = seg_lens[0]
    seg_lens = np.array([first_len] + list(seg_lens[2:]))
    seg_ctr_pts = seg_ctr_pts[1:-1]  # remove end points again

    conductivity = 0.3

    elec_pos = (10, 150, 0)
    target_vals = {"psa": list(), "lsa": list()}
    for seg_idx in range(sec.nseg):
        # PSA: distance to middle segment == electrode x-position
        var_r_psa = np.sqrt(
            elec_pos[0] ** 2 + (elec_pos[1] - seg_ctr_pts[seg_idx]) ** 2
        )
        target_vals["psa"].append(1000 / (4.0 * np.pi * conductivity * var_r_psa))

        # LSA: calculate L and H variables relative to segment endpoints
        var_l = elec_pos[1] - (seg_ctr_pts[seg_idx] - seg_lens[seg_idx])
        var_h = elec_pos[1] - (seg_ctr_pts[seg_idx] + seg_lens[seg_idx])
        var_r_lsa = elec_pos[0]  # just use the axial distance
        target_vals["lsa"].append(
            1000
            * np.log(
                np.abs(
                    (np.sqrt(var_h**2 + var_r_lsa**2) - var_h)
                    / (np.sqrt(var_l**2 + var_r_lsa**2) - var_l)
                )
            )
            / (4.0 * np.pi * conductivity * 2 * seg_lens[seg_idx])
        )

    for method in ["psa", "lsa"]:
        res = _transfer_resistance(sec, elec_pos, conductivity, method)
        assert_allclose(res, target_vals[method], rtol=1e-12, atol=0.0)


@requires_mpi4py
@requires_psutil
@pytest.mark.uses_mpi
def test_extracellular_backends(run_hnn_core_fixture):
    """Test extracellular outputs across backends."""
    # calculation of CSD requires >=4 electrode contacts
    electrode_array = {"arr1": [(2, 2, 400), (2, 2, 600), (2, 2, 800), (2, 2, 1000)]}
    _, joblib_net = run_hnn_core_fixture(
        backend="joblib",
        n_jobs=1,
        reduced=True,
        record_isec="soma",
        record_vsec="soma",
        record_ca="soma",
        electrode_array=electrode_array,
    )
    _, mpi_net = run_hnn_core_fixture(
        backend="mpi",
        n_procs=2,
        reduced=True,
        record_isec="soma",
        record_vsec="soma",
        record_ca="soma",
        electrode_array=electrode_array,
    )

    assert (
        len(electrode_array["arr1"])
        == len(joblib_net.rec_arrays["arr1"].positions)
        == len(mpi_net.rec_arrays["arr1"].positions)
    )
    assert (
        len(joblib_net.rec_arrays["arr1"]) == len(mpi_net.rec_arrays["arr1"]) == 2
    )  # length == n.o. trials

    # reduced simulation has n_trials=2
    # trial_idx, n_trials = 0, 2
    for tr_idx, el_idx in zip([0, 1], [0, 1]):
        assert_allclose(
            joblib_net.rec_arrays["arr1"]._data[tr_idx][el_idx],
            mpi_net.rec_arrays["arr1"]._data[tr_idx][el_idx],
        )

    assert isinstance(joblib_net.rec_arrays["arr1"].voltages, np.ndarray)
    assert_array_equal(
        joblib_net.rec_arrays["arr1"].voltages.shape,
        [
            len(joblib_net.rec_arrays["arr1"]._data),
            len(joblib_net.rec_arrays["arr1"]._data[0]),
            len(joblib_net.rec_arrays["arr1"]._data[0][0]),
        ],
    )

    # make sure sampling rate is fixed (raises RuntimeError if not)
    _ = joblib_net.rec_arrays["arr1"].sfreq
    # check plotting works
    joblib_net.rec_arrays["arr1"].plot_lfp(show=False)
    joblib_net.rec_arrays["arr1"].plot_csd(show=False)

    plt.close("all")


def test_rec_array_calculation():
    """Test LFP/CSD calculation."""
    hnn_core_root = op.dirname(hnn_core.__file__)
    params_fname = op.join(hnn_core_root, "param", "default.json")
    params = read_params(params_fname)
    params.update({"t_evprox_1": 7, "t_evdist_1": 17})
    net = jones_2009_model(params, mesh_shape=(3, 3), add_drives_from_params=True)

    # one electrode inside, one above the active elements of the network,
    # and two more to allow calculation of CSD (2nd spatial derivative)
    electrode_pos = [(1, 2, 1000), (2, 3, 3000), (3, 4, 5000), (4, 5, 7000)]
    net.add_electrode_array("arr1", electrode_pos)
    _ = simulate_dipole(net, tstop=5, n_trials=1)

    # test accessing simulated voltages
    assert (
        len(net.rec_arrays["arr1"]) == len(net.rec_arrays["arr1"].voltages) == 1
    )  # n_trials
    assert len(net.rec_arrays["arr1"].voltages[0]) == 4  # n_contacts
    assert len(net.rec_arrays["arr1"].voltages[0][0]) == len(
        net.rec_arrays["arr1"].times
    )

    # test dimensionality of LFP and CSD matrices
    lfp_data = net.rec_arrays["arr1"].voltages[0]
    csd_data = calculate_csd2d(lfp_data)
    assert lfp_data.shape == csd_data.shape

    # ensure copy drops data (but retains electrode position information etc.)
    net_copy = net.copy()
    assert isinstance(net_copy.rec_arrays["arr1"], ExtracellularArray)
    assert len(net_copy.rec_arrays["arr1"].voltages) == 0

    assert isinstance(net.rec_arrays["arr1"].voltages, np.ndarray)
    assert isinstance(net.rec_arrays["arr1"].times, np.ndarray)

    # using the same electrode positions, but a different method: LSA
    net.add_electrode_array("arr2", electrode_pos, method="lsa")

    # make sure no sinister segfaults are triggered when running mult. trials
    n_trials = 5  # NB 5 trials!
    _ = simulate_dipole(net, tstop=5, n_trials=n_trials)

    # simulate_dipole is run twice above, first 1 then 5 trials.
    # Make sure that previous results are discarded on each run
    assert len(net.rec_arrays["arr1"]._data) == n_trials

    for trial_idx in range(n_trials):
        # LSA and PSA should agree far away (second electrode)
        assert_allclose(
            net.rec_arrays["arr1"]._data[trial_idx][1],
            net.rec_arrays["arr2"]._data[trial_idx][1],
            rtol=1e-3,
            atol=1e-3,
        )


def test_extracellular_viz():
    """Test if deprecation warning is raised in plot_laminar_lfp."""
    hnn_core_root = op.dirname(hnn_core.__file__)
    params_fname = op.join(hnn_core_root, "param", "default.json")
    params = read_params(params_fname)
    params.update({"t_evprox_1": 7, "t_evdist_1": 17})
    net = jones_2009_model(params, mesh_shape=(3, 3), add_drives_from_params=True)

    # one electrode inside, one above the active elements of the network,
    # and two more to allow calculation of CSD (2nd spatial derivative)
    electrode_pos = [(1, 2, 1000), (2, 3, 3000), (3, 4, 5000), (4, 5, 7000)]
    net.add_electrode_array("arr1", electrode_pos)
    _ = simulate_dipole(net, tstop=5, n_trials=1)

    with pytest.deprecated_call():
        net.rec_arrays["arr1"].plot_lfp(show=False, tmin=10, tmax=100)
    with pytest.raises(
        RuntimeError,
        match='Please use sink = "b" or '
        'sink = "r". Only colormap "jet" is supported '
        "for CSD.",
    ):
        net.rec_arrays["arr1"].plot_csd(show=False, sink="g")



# Authors: Mainak Jas <mainakjas@gmail.com>

import os.path as op
import numpy as np
import pytest

import hnn_core
from hnn_core import read_params, jones_2009_model, simulate_dipole
from hnn_core.optimization.optimize_evoked import (
    _consolidate_chunks,
    _split_by_evinput,
    _generate_weights,
    _get_drive_params,
    optimize_evoked,
)


def test_consolidate_chunks():
    """Test consolidation of chunks."""
    inputs = {
        "ev1": {
            "start": 5,
            "end": 25,
            "ranges": {"initial": 1e-10, "minval": 1e-11, "maxval": 1e-9},
            "opt_end": 90,
            "weights": np.array([5.0, 10.0]),
        },
        "ev2": {
            "start": 100,
            "end": 120,
            "ranges": {"initial": 1e-10, "minval": 1e-11, "maxval": 1e-9},
            "opt_end": 170,
            "weights": np.array([10.0, 5.0]),
        },
    }
    chunks = _consolidate_chunks(inputs)
    assert len(chunks) == len(inputs) + 1  # extra last chunk??
    assert chunks[-1]["opt_end"] == inputs["ev2"]["opt_end"]
    assert chunks[-1]["inputs"] == ["ev1", "ev2"]
    assert isinstance(chunks, list)

    # overlapping chunks
    inputs["ev1"]["end"] = 110
    chunks = _consolidate_chunks(inputs)
    assert len(chunks) == 1
    assert chunks[0]["start"] == inputs["ev1"]["start"]
    assert chunks[0]["end"] == inputs["ev2"]["end"]
    assert np.allclose(
        chunks[0]["weights"],
        (inputs["ev1"]["weights"] + inputs["ev2"]["weights"]) / 2.0,
    )


def test_split_by_evinput():
    """Test splitting evoked input."""
    drive_names = ["ev_drive_1", "ev_drive_2"]
    drive_dynamics = [{"mu": 5.0, "sigma": 0.1}, {"mu": 10.0, "sigma": 0.2}]
    drive_syn_weights = [{"ampa_L2_pyramidal": 1.0}, {"nmda_L5_basket": 2.0}]
    tstop = 20.0
    dt = 0.025

    timing_range_multiplier = 3.0
    sigma_range_multiplier = 50.0
    synweight_range_multiplier = 500.0
    decay_multiplier = 1.6
    evinput_params = _split_by_evinput(
        drive_names,
        drive_dynamics,
        drive_syn_weights,
        tstop,
        sigma_range_multiplier,
        timing_range_multiplier,
        synweight_range_multiplier,
    )
    assert list(evinput_params.keys()) == drive_names
    for evinput in evinput_params.values():
        assert list(evinput.keys()) == ["mean", "sigma", "ranges", "start", "end"]

    evinput_params = _generate_weights(evinput_params, tstop, dt, decay_multiplier)
    for evinput in evinput_params.values():
        assert list(evinput.keys()) == [
            "ranges",
            "start",
            "end",
            "weights",
            "opt_start",
            "opt_end",
        ]


def test_optimize_evoked():
    """Test running the full routine in a reduced network."""
    hnn_core_root = op.dirname(hnn_core.__file__)
    params_fname = op.join(hnn_core_root, "param", "default.json")
    params = read_params(params_fname)

    tstop = 10.0
    n_trials = 1

    # simulate a dipole to establish ground-truth drive parameters
    mu_orig = 6.0
    params.update(
        {
            "t_evprox_1": mu_orig,
            "sigma_t_evprox_1": 2.0,
            "t_evdist_1": mu_orig + 2,
            "sigma_t_evdist_1": 2.0,
        }
    )
    net_orig = jones_2009_model(params, add_drives_from_params=True, mesh_shape=(3, 3))
    del net_orig.external_drives["evprox2"]
    dpl_orig = simulate_dipole(net_orig, tstop=tstop, n_trials=n_trials)[0]

    # simulate a dipole with a time-shifted drive
    mu_offset = 4.0
    params.update(
        {
            "t_evprox_1": mu_offset,
            "sigma_t_evprox_1": 2.0,
            "t_evdist_1": mu_offset + 2,
            "sigma_t_evdist_1": 2.0,
        }
    )
    net_offset = jones_2009_model(
        params, add_drives_from_params=True, mesh_shape=(3, 3)
    )
    del net_offset.external_drives["evprox2"]
    dpl_offset = simulate_dipole(net_offset, tstop=tstop, n_trials=n_trials)[0]
    # get drive params from the pre-optimization Network instance
    _, _, drive_static_params_orig = _get_drive_params(net_offset, ["evprox1"])

    with pytest.raises(
        ValueError, match="The current Network instance lacks any evoked drives"
    ):
        net_empty = net_offset.copy()
        del net_empty.external_drives["evprox1"]
        del net_empty.external_drives["evdist1"]
        net_opt = optimize_evoked(
            net_empty,
            tstop=tstop,
            n_trials=n_trials,
            target_dpl=dpl_orig,
            initial_dpl=dpl_offset,
            maxiter=10,
        )

    with pytest.raises(
        ValueError, match="The drives selected to be optimized are not evoked drives"
    ):
        net_test_bursty = net_offset.copy()
        which_drives = ["bursty1"]
        net_opt = optimize_evoked(
            net_test_bursty,
            tstop=tstop,
            n_trials=n_trials,
            target_dpl=dpl_orig,
            initial_dpl=dpl_offset,
            which_drives=which_drives,
            maxiter=10,
        )

    which_drives = ["evprox1"]  # drive selected to optimize
    maxiter = 12
    # try without returning iteration RMSE first
    net_opt = optimize_evoked(
        net_offset,
        tstop=tstop,
        n_trials=n_trials,
        target_dpl=dpl_orig,
        initial_dpl=dpl_offset,
        timing_range_multiplier=3.0,
        sigma_range_multiplier=50.0,
        synweight_range_multiplier=500.0,
        maxiter=maxiter,
        which_drives=which_drives,
        return_rmse=False,
    )
    net_opt, rmse = optimize_evoked(
        net_offset,
        tstop=tstop,
        n_trials=n_trials,
        target_dpl=dpl_orig,
        initial_dpl=dpl_offset,
        timing_range_multiplier=3.0,
        sigma_range_multiplier=50.0,
        synweight_range_multiplier=500.0,
        maxiter=maxiter,
        which_drives=which_drives,
        return_rmse=True,
    )

    # the number of returned rmse values should be the same as maxiter
    assert len(rmse) <= maxiter

    # the returned rmse values should be positive
    assert all(vals > 0 for vals in rmse)

    # the names of drives should be preserved during optimization
    assert net_offset.external_drives.keys() == net_opt.external_drives.keys()

    drive_dynamics_opt, drive_syn_weights_opt, drive_static_params_opt = (
        _get_drive_params(net_opt, ["evprox1"])
    )

    # ensure that params corresponding to only one evoked drive are discovered
    assert (
        len(drive_dynamics_opt)
        == len(drive_syn_weights_opt)
        == len(drive_static_params_opt)
        == 1
    )

    # static drive params should remain constant
    assert drive_static_params_opt == drive_static_params_orig

    # ensure that only the drive that we wanted to optimize over changed
    (
        drive_evdist1_dynamics_offset,
        drive_evdist1_syn_weights_offset,
        drive_static_params_offset,
    ) = _get_drive_params(net_offset, ["evdist1"])
    (
        drive_evdist1_dynamics_opt,
        drive_evdist1_syn_weights_opt,
        drive_static_params_opt,
    ) = _get_drive_params(net_opt, ["evdist1"])

    # assert that evdist1 did NOT change
    assert drive_evdist1_dynamics_opt == drive_evdist1_dynamics_offset
    assert drive_evdist1_syn_weights_opt == drive_evdist1_syn_weights_offset
